# Copyright 2010-2017, The University of Melbourne
# Copyright 2010-2017, Brian May
#
# This file is part of Karaage.
#
# Karaage is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Karaage is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Karaage  If not, see <http://www.gnu.org/licenses/>.

import django_tables2 as tables
import six
from django.contrib import messages
from django.db.models import Q
from django.http import HttpResponseForbidden, HttpResponseRedirect
from django.shortcuts import get_object_or_404, render
from django.urls import reverse

import karaage.common as util
from karaage.common.decorators import admin_required, login_required
from karaage.machines.models import Account
from karaage.people.models import Person
from karaage.projects.forms import AddPersonForm, ProjectForm, UserProjectForm
from karaage.projects.models import Project
from karaage.projects.tables import ProjectFilter, ProjectTable
from karaage.projects.utils import (
    add_user_to_project,
    get_new_pid,
    remove_user_from_project,
)


@login_required
def profile_projects(request):
    config = tables.RequestConfig(request, paginate={"per_page": 5})

    person = request.user
    project_list = person.projects.all()
    project_list = ProjectTable(project_list, prefix="mine-")
    config.configure(project_list)

    delegate_project_list = Project.objects.filter(
        institute__delegates=person, is_active=True)
    delegate_project_list = ProjectTable(
        delegate_project_list, prefix="delegate-")
    config.configure(delegate_project_list)

    leader_project_list = Project.objects.filter(
        leaders=person, is_active=True)
    leader_project_list = ProjectTable(
        leader_project_list, prefix="leader-")
    config.configure(leader_project_list)

    return render(
        template_name='karaage/projects/profile_projects.html',
        context={
            'person': person,
            'project_list': project_list,
            'delegate_project_list': delegate_project_list,
            'leader_project_list': leader_project_list
        },
        request=request)


@login_required
def add_edit_project(request, project_id=None):

    if project_id is None:
        project = None
        old_pid = None
        flag = 1
    else:
        project = get_object_or_404(Project, id=project_id)
        old_pid = project.pid
        flag = 2

    if util.is_admin(request):
        form = ProjectForm(instance=project, data=request.POST or None)
    else:
        if not project.can_edit(request):
            return HttpResponseForbidden('<h1>Access Denied</h1>')
        form = UserProjectForm(instance=project, data=request.POST or None)

    if request.method == 'POST':
        if form.is_valid():
            project = form.save(commit=False)
            if project_id is not None:
                # if project is being edited, project_id cannot change.
                project.pid = old_pid
            elif not project.pid:
                # if project was being created, did the user give a project_id
                # we should use? If not, then we have to generate one
                # ourselves.
                project.pid = get_new_pid(project.institute)
            project.save()
            approved_by = request.user
            project.activate(approved_by)
            form.save_m2m()
            if flag == 1:
                messages.success(
                    request, "Project '%s' created succesfully" % project)
            else:
                messages.success(
                    request, "Project '%s' edited succesfully" % project)

            return HttpResponseRedirect(project.get_absolute_url())

    return render(
        template_name='karaage/projects/project_form.html',
        context=locals(),
        request=request)


@admin_required
def undelete_project(request, project_id):

    project = get_object_or_404(Project, id=project_id)

    if request.method == 'POST':
        undeleted_by = request.user
        project.activate(undeleted_by)
        messages.success(
            request, "Project '%s' undeleted succesfully" % project)
        return HttpResponseRedirect(project.get_absolute_url())

    return render(
        template_name='karaage/projects/project_confirm_undelete.html',
        context={'project': project},
        request=request)


@admin_required
def delete_project(request, project_id):

    project = get_object_or_404(Project, id=project_id)

    query = Account.objects.filter(
        date_deleted__isnull=True, default_project=project)

    error = None
    if query.count() > 0:
        error = "There are accounts that use this project " \
            "as the default_project."

    elif request.method == 'POST':
        deleted_by = request.user
        project.deactivate(deleted_by)
        messages.success(request, "Project '%s' deleted succesfully" % project)
        return HttpResponseRedirect(project.get_absolute_url())

    del query

    return render(
        template_name='karaage/projects/project_confirm_delete.html',
        context={'project': project, 'error': error},
        request=request)


@login_required
def project_detail(request, project_id):

    project = get_object_or_404(Project, id=project_id)

    if not project.can_view(request):
        return HttpResponseForbidden('<h1>Access Denied</h1>')

    form = AddPersonForm(request.POST or None)
    if request.method == 'POST':
        # Post means adding a user to this project
        if form.is_valid():
            person = form.cleaned_data['person']
            add_user_to_project(person, project)
            messages.success(
                request,
                "User '%s' was added to %s succesfully" % (person, project))
            return HttpResponseRedirect(project.get_absolute_url())

    return render(
        template_name='karaage/projects/project_detail.html',
        context={
            'project': project,
            'form': form,
            'can_edit': project.can_edit(request)
        },
        request=request)


@admin_required
def project_verbose(request, project_id):
    project = get_object_or_404(Project, id=project_id)

    from karaage.datastores import get_project_details
    project_details = get_project_details(project)

    return render(
        template_name='karaage/projects/project_verbose.html',
        context=locals(),
        request=request)


@login_required
def project_list(request, queryset=None):

    if queryset is None:
        queryset = Project.objects.all()

    if not util.is_admin(request):
        queryset = queryset.filter(
            Q(leaders=request.user, is_active=True)
            | Q(institute__delegates=request.user, is_active=True)
            | Q(group__members=request.user, is_active=True)).distinct()

    queryset = queryset.select_related()

    q_filter = ProjectFilter(request.GET, queryset=queryset)
    table = ProjectTable(q_filter.qs)
    tables.RequestConfig(request).configure(table)

    spec = []
    for name, value in six.iteritems(q_filter.form.cleaned_data):
        if value is not None and value != "":
            name = name.replace('_', ' ').capitalize()
            spec.append((name, value))

    return render(
        template_name='karaage/projects/project_list.html',
        context={
            'table': table,
            'filter': q_filter,
            'spec': spec,
            'title': "Project list",
        },
        request=request)


@login_required
def remove_user(request, project_id, username):

    project = get_object_or_404(Project, id=project_id)
    person = get_object_or_404(Person, username=username)

    if not project.can_edit(request):
        return HttpResponseForbidden('<h1>Access Denied</h1>')

    query = person.account_set.filter(
        date_deleted__isnull=True, default_project=project)

    error = None
    if query.count() > 0:
        error = "The person has accounts that use this project " \
            "as the default_project."

    elif request.method == 'POST':
        remove_user_from_project(person, project)
        messages.success(
            request,
            "User '%s' removed succesfully from project %s"
            % (person, project.pid))
        return HttpResponseRedirect(project.get_absolute_url())

    del query

    return render(
        template_name='karaage/projects/remove_user_confirm.html',
        context={'project': project, 'person': person, 'error': error, },
        request=request)


@login_required
def grant_leader(request, project_id, username):
    project = get_object_or_404(Project, id=project_id)
    person = get_object_or_404(Person, username=username)

    if not project.can_edit(request):
        return HttpResponseForbidden('<h1>Access Denied</h1>')

    if project.group.members.filter(pk=person.pk).count() == 0:
        return HttpResponseForbidden('<h1>Access Denied</h1>')

    if request.method == 'POST':
        if project.leaders.filter(pk=person.pk).count() == 0:
            project.leaders.add(person)
            messages.success(
                request,
                "User '%s' granted leader rights for project %s"
                % (person, project.pid))
        return HttpResponseRedirect(project.get_absolute_url())

    return render(
        template_name='karaage/projects/grant_leader.html',
        context={'project': project, 'person': person, },
        request=request)


@login_required
def revoke_leader(request, project_id, username):
    project = get_object_or_404(Project, id=project_id)
    person = get_object_or_404(Person, username=username)

    if not project.can_edit(request):
        return HttpResponseForbidden('<h1>Access Denied</h1>')

    error = None
    if request.user == person:
        error = "Cannot revoke self."

    elif project.leaders.exclude(pk=person.pk).count() == 0:
        error = "Cannot revoke last project leader."

    elif request.method == 'POST':
        if project.leaders.filter(pk=person.pk).count() > 0:
            project.leaders.remove(person)
            messages.success(
                request,
                "User '%s' revoked leader rights for project %s"
                % (person, project.pid))
        return HttpResponseRedirect(project.get_absolute_url())

    return render(
        template_name='karaage/projects/revoke_leader.html',
        context={'project': project, 'person': person, 'error': error},
        request=request)


@admin_required
def no_users(request):

    project_ids = []
    for p in Project.active.all():
        if p.group.members.count() == 0:
            project_ids.append(p.id)

    return project_list(request, Project.objects.filter(id__in=project_ids))


@admin_required
def project_logs(request, project_id):
    obj = get_object_or_404(Project, id=project_id)
    breadcrumbs = [
        ("Projects", reverse("kg_project_list")),
        (six.text_type(obj.pid), reverse("kg_project_detail", args=[obj.id]))
    ]
    return util.log_list(request, breadcrumbs, obj)


@admin_required
def add_comment(request, project_id):
    obj = get_object_or_404(Project, id=project_id)
    breadcrumbs = [
        ("Projects", reverse("kg_project_list")),
        (six.text_type(obj.pid), reverse("kg_project_detail", args=[obj.id]))
    ]
    return util.add_comment(request, breadcrumbs, obj)



# -*- coding: utf-8 -*-
# Generated by Django 1.11.10 on 2018-03-02 17:35
from __future__ import unicode_literals

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('karaage', '0005_auto_20171215_1831'),
    ]

    operations = [
        migrations.AlterField(
            model_name='account',
            name='default_project',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, to='karaage.Project'),
        ),
        migrations.AlterField(
            model_name='institute',
            name='group',
            field=models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, to='karaage.Group'),
        ),
        migrations.AlterField(
            model_name='logentry',
            name='user',
            field=models.ForeignKey(null=True, on_delete=django.db.models.deletion.SET_NULL, to=settings.AUTH_USER_MODEL),
        ),
        migrations.AlterField(
            model_name='person',
            name='approved_by',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='user_approver', to=settings.AUTH_USER_MODEL),
        ),
        migrations.AlterField(
            model_name='person',
            name='deleted_by',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='user_deletor', to=settings.AUTH_USER_MODEL),
        ),
        migrations.AlterField(
            model_name='project',
            name='approved_by',
            field=models.ForeignKey(blank=True, editable=False, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='project_approver', to=settings.AUTH_USER_MODEL),
        ),
        migrations.AlterField(
            model_name='project',
            name='deleted_by',
            field=models.ForeignKey(blank=True, editable=False, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='project_deletor', to=settings.AUTH_USER_MODEL),
        ),
        migrations.AlterField(
            model_name='project',
            name='group',
            field=models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, to='karaage.Group'),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.20 on 2019-03-12 08:05
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('kgapplications', '0005_auto_20180302_1808'),
    ]

    operations = [
        migrations.AlterField(
            model_name='applicant',
            name='saml_id',
            field=models.CharField(blank=True, editable=False, max_length=200, null=True),
        ),
    ]



# Copyright 2010-2017, The University of Melbourne
# Copyright 2010-2017, Brian May
#
# This file is part of Karaage.
#
# Karaage is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Karaage is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Karaage  If not, see <http://www.gnu.org/licenses/>.

import unittest

import mock
import pytest
from django.core import exceptions as django_exceptions

from karaage.people.models import Group
from karaage.tests.fixtures import (
    GroupFactory,
    InstituteFactory,
    simple_account,
)
from karaage.tests.unit import UnitTestCase


@pytest.mark.django_db
class InstituteTestCase(UnitTestCase):

    def test_add(self):
        InstituteFactory(name='TestInstitute54')

    def test_add_existing_name(self):
        group, _ = Group.objects.get_or_create(name='testinstitute27')
        institute = InstituteFactory(
            name='Test Institute 27', group=group)
        self.assertEqual(
            institute.group.name,
            'testinstitute27')
        self.assertEqual(
            institute.group.name,
            institute.name.lower().replace(' ', ''))

    @unittest.skip("broken with mysql/postgresql")
    def test_username(self):
        assert_raises = self.assertRaises(django_exceptions.ValidationError)

        # Max length
        institution = InstituteFactory(name="a" * 255)
        institution.full_clean()

        # Name is too long
        institution = InstituteFactory(name="a" * 256)
        with assert_raises:
            institution.full_clean()

    def test_change_group(self):
        """Check that when changing an institutes group, old accounts are
        removed from the institute and new ones are added.

        """
        account1 = simple_account()
        group1 = GroupFactory()

        # Test initial creation of the institute
        self.resetDatastore()
        institute = InstituteFactory(group=group1)
        self.assertEqual(
            self.datastore.method_calls,
            [mock.call.save_institute(institute)])

        # Test setting up initial group for institute
        self.resetDatastore()
        group1.save()
        group1.add_person(account1.person)
        self.assertEqual(
            self.datastore.method_calls,
            [mock.call.save_group(group1),
             mock.call.add_account_to_group(account1, group1),
             mock.call.add_account_to_institute(account1, institute)])

        # Test changing an existing institutions group
        account2 = simple_account(institute=institute)
        self.resetDatastore()
        group2 = GroupFactory()
        group2.add_person(account2.person)
        group2.save()
        institute.group = group2
        institute.save()
        self.assertEqual(
            self.datastore.method_calls,
            [mock.call.save_group(group2),
             mock.call.add_account_to_group(account2, group2),
             mock.call.save_group(group2),
             mock.call.save_institute(institute),
             # old accounts are removed
             mock.call.remove_account_from_institute(account1, institute),
             # new accounts are added
             mock.call.add_account_to_institute(account2, institute)])

        # Test deleting institute
        self.resetDatastore()
        institute.delete()
        self.assertEqual(
            self.datastore.method_calls,
            [mock.call.remove_account_from_institute(account2, institute),
             mock.call.delete_institute(institute)])



# Copyright 2010-2017, The University of Melbourne
# Copyright 2010-2017, Brian May
#
# This file is part of Karaage.
#
# Karaage is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Karaage is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Karaage  If not, see <http://www.gnu.org/licenses/>.

import django_filters
import django_tables2 as tables
from django_tables2.utils import A

from .models import Application
from .views.base import get_state_machine


class ApplicationFilter(django_filters.FilterSet):

    class Meta:
        model = Application
        fields = ('secret_token',)


class ApplicationTable(tables.Table):
    id = tables.LinkColumn(
        'kg_application_detail', args=[A('pk')], verbose_name="ID")
    action = tables.Column(
        empty_values=(), order_by=('_class'))
    applicant = tables.Column(
        linkify=True, order_by=('content_type', 'object_id'))

    def render_action(self, record):
        return record.get_object().info()

    def render_state(self, record):
        state_machine = get_state_machine(record)
        state = state_machine.get_state(record)
        return state.name

    class Meta:
        model = Application
        fields = ('id', 'action', 'applicant', 'state', 'expires')
        empty_text = "No items"



# Copyright 2010-2017, The University of Melbourne
# Copyright 2010-2017, Brian May
#
# This file is part of Karaage.
#
# Karaage is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Karaage is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Karaage  If not, see <http://www.gnu.org/licenses/>.

import datetime

from django.conf import settings
from django.contrib.auth.hashers import make_password

import karaage.institutes.models
import karaage.machines.models
import karaage.people.models
import karaage.projects.models
from karaage.projects.utils import add_user_to_project


try:
    from factory.django import DjangoModelFactory
    from factory.fuzzy import FuzzyText, FuzzyChoice
    import factory
except ImportError:
    raise ImportError(
        "factory_boy is required, "
        "either install from a package or using \'pip install -e .[tests]\'")


def fuzzy_lower_text(*args, **kwargs):
    if 'chars' not in kwargs:
        kwargs['chars'] = 'abcdefghijklmnopqrstuvwxyz'
    return FuzzyText(*args, **kwargs)


class GroupFactory(DjangoModelFactory):
    name = fuzzy_lower_text(prefix='group-')

    class Meta:
        model = karaage.people.models.Group
        django_get_or_create = ('name',)


class PersonFactory(DjangoModelFactory):
    username = fuzzy_lower_text(prefix='person-')
    password = make_password('test')
    full_name = factory.LazyAttribute(lambda a: a.username.title()[:50])
    short_name = factory.LazyAttribute(lambda a: a.username.title()[:30])
    email = factory.LazyAttribute(
        lambda a: '{0}@example.com'.format(a.username[:62]).lower())
    institute = factory.LazyAttribute(
        lambda a: InstituteFactory(
            name=fuzzy_lower_text(prefix='%s-inst-' % a.username)))

    class Meta:
        model = karaage.people.models.Person
        django_get_or_create = ('username',)


class InstituteFactory(DjangoModelFactory):
    name = fuzzy_lower_text(prefix='inst-')
    group = factory.SubFactory(GroupFactory)

    class Meta:
        model = karaage.institutes.models.Institute
        django_get_or_create = ('name',)


class ProjectFactory(DjangoModelFactory):
    pid = fuzzy_lower_text(prefix='proj-')
    name = factory.LazyAttribute(lambda a: a.pid.title()[:200])
    institute = factory.SubFactory(InstituteFactory)
    is_approved = True
    approved_by = factory.LazyAttribute(
        lambda a: PersonFactory(
            username=fuzzy_lower_text(prefix='%s-acc-' % a.pid)))
    is_active = True

    class Meta:
        model = karaage.projects.models.Project


class AccountFactory(DjangoModelFactory):
    username = fuzzy_lower_text(prefix='account-')
    foreign_id = FuzzyText()
    person = factory.SubFactory(PersonFactory)
    date_created = factory.LazyAttribute(lambda a: datetime.datetime.today())
    default_project = factory.SubFactory(ProjectFactory)
    shell = FuzzyChoice([s[0] for s in settings.SHELLS])

    class Meta:
        model = karaage.machines.models.Account
        django_get_or_create = ('person', 'username')


def simple_account(institute=None):
    if not institute:
        institute = InstituteFactory()
    person = PersonFactory(institute=institute)
    project = ProjectFactory(pid=fuzzy_lower_text(prefix='proj-default-'),
                             institute=institute,
                             approved_by=person)
    account = AccountFactory(person=person,
                             default_project=project)
    add_user_to_project(account.person, project)
    return account



# Copyright 2010-2017, The University of Melbourne
# Copyright 2010-2017, Brian May
#
# This file is part of Karaage.
#
# Karaage is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Karaage is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Karaage  If not, see <http://www.gnu.org/licenses/>.
import sys
import re
from os.path import abspath, dirname, join

sys.path.insert(0, abspath('../..'))
import karaage  # NOQA


sys.path.append(abspath(join(dirname(dirname(__file__)), "ext")))

html_translator_class = "djangodocs.KaraageHTMLTranslator"

extensions = []
extensions += ['sphinx.ext.todo']
extensions += ["djangodocs"]
extensions += ["sphinx.ext.intersphinx"]
intersphinx_mapping = {
    'django': (
        'https://docs.djangoproject.com/en/1.9/',
        'https://docs.djangoproject.com/en/1.9/_objects/'),
}

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#

# The full version, including alpha/beta/rc tags.
release = karaage.__version__
# The short X.Y version.
version = '.'.join(release.split('.')[:2])


def guess_next_version(tag_version):
    version = str(tag_version)
    if '.dev' in version:
        version, tail = version.rsplit('.dev', 1)

    prefix, tail = re.match('(.*?)(\d+)$', version).groups()
    return '%s%d' % (prefix, int(tail) + 1)


karaage_next_version = guess_next_version(version)



# Copyright 2010-2017, The University of Melbourne
# Copyright 2010-2017, Brian May
#
# This file is part of Karaage.
#
# Karaage is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Karaage is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Karaage  If not, see <http://www.gnu.org/licenses/>.

"""
Holds various constants
"""

TITLES = (
    ('', ''),
    ('Mr', 'Mr'),
    ('Mrs', 'Mrs'),
    ('Miss', 'Miss'),
    ('Ms', 'Ms'),
    ('Dr', 'Dr'),
    ('Prof', 'Prof'),
    ('A/Prof', 'A/Prof'),
)

DATE_FORMATS = [
    # '2006-10-25'
    '%Y-%m-%d',

    # '25-10-06', '25/10/06', '25/10/06'
    '%d-%m-%y', '%d/%m/%y', '%d/%m/%y',

    # '25-10-2006', '25/10/2006', '25/10/2006'
    '%d-%m-%Y', '%d/%m/%Y', '%d/%m/%Y',

    # 'Oct 25 2006', 'Oct 25, 2006'
    '%b %d %Y', '%b %d, %Y',

    # '25 Oct 2006', '25 Oct, 2006
    '%d %b %Y', '%d %b, %Y',

    # 'October 25 2006', 'October 25, 2006'
    '%B %d %Y', '%B %d, %Y',

    # '25 October 2006', '25 October, 2006'
    '%d %B %Y', '%d %B, %Y',
]

STATES = (
    ('', '--------'),
    ('ACT', 'ACT'),
    ('NSW', 'New South Wales'),
    ('NT', 'Northern Territory'),
    ('QLD', 'Queensland'),
    ('SA', 'South Australia'),
    ('TAS', 'Tasmania'),
    ('VIC', 'Victoria'),
    ('WA', 'Western Australia'),
)

COUNTRIES = (
    ('AU', 'Australia'),
    ('NZ', 'New Zealand'),
    ('GB', 'United Kingdom'),
    ('DE', 'Germany'),
    ('US', 'United States'),
    ('', '--------------------------------------'),
    ('AD', 'Andorra'),
    ('AE', 'United Arab Emirates'),
    ('AF', 'Afghanistan'),
    ('AG', 'Antigua and Barbuda'),
    ('AI', 'Anguilla'),
    ('AL', 'Albania'),
    ('AM', 'Armenia'),
    ('AN', 'Netherlands Antilles'),
    ('AO', 'Angola'),
    ('AQ', 'Antarctica'),
    ('AR', 'Argentina'),
    ('AS', 'American Samoa'),
    ('AT', 'Austria'),
    ('AW', 'Aruba'),
    ('AX', 'Aland Islands'),
    ('AZ', 'Azerbaijan'),
    ('BA', 'Bosnia and Herzegovina'),
    ('BB', 'Barbados'),
    ('BD', 'Bangladesh'),
    ('BE', 'Belgium'),
    ('BF', 'Burkina Faso'),
    ('BG', 'Bulgaria'),
    ('BH', 'Bahrain'),
    ('BI', 'Burundi'),
    ('BJ', 'Benin'),
    ('BM', 'Bermuda'),
    ('BN', 'Brunei Darussalam'),
    ('BO', 'Bolivia'),
    ('BR', 'Brazil'),
    ('BS', 'Bahamas'),
    ('BT', 'Bhutan'),
    ('BV', 'Bouvet Island'),
    ('BW', 'Botswana'),
    ('BY', 'Belarus'),
    ('BZ', 'Belize'),
    ('CA', 'Canada'),
    ('CC', 'Cocos (Keeling) Islands'),
    ('CD', 'Congo'),
    ('CF', 'Central African Republic'),
    ('CG', 'Congo'),
    ('CH', 'Switzerland'),
    ('CI', "Cote d'Ivoire"),
    ('CK', 'Cook Islands'),
    ('CL', 'Chile'),
    ('CM', 'Cameroon'),
    ('CN', 'China'),
    ('CO', 'Colombia'),
    ('CR', 'Costa Rica'),
    ('CU', 'Cuba'),
    ('CV', 'Cape Verde'),
    ('CX', 'Christmas Island'),
    ('CY', 'Cyprus'),
    ('CZ', 'Czech Republic'),
    ('DJ', 'Djibouti'),
    ('DK', 'Denmark'),
    ('DM', 'Dominica'),
    ('DO', 'Dominican Republic'),
    ('DZ', 'Algeria'),
    ('EC', 'Ecuador'),
    ('EE', 'Estonia'),
    ('EG', 'Egypt'),
    ('EH', 'Western Sahara'),
    ('ER', 'Eritrea'),
    ('ES', 'Spain'),
    ('ET', 'Ethiopia'),
    ('FI', 'Finland'),
    ('FJ', 'Fiji'),
    ('FK', 'Falkland Islands'),
    ('FM', 'Micronesia'),
    ('FO', 'Faroe Islands'),
    ('FR', 'France'),
    ('GA', 'Gabon'),
    ('GD', 'Grenada'),
    ('GE', 'Georgia'),
    ('GF', 'French Guiana'),
    ('GG', 'Guernsey'),
    ('GH', 'Ghana'),
    ('GI', 'Gibraltar'),
    ('GL', 'Greenland'),
    ('GM', 'Gambia'),
    ('GN', 'Guinea'),
    ('GP', 'Guadeloupe'),
    ('GQ', 'Equatorial Guinea'),
    ('GR', 'Greece'),
    ('GS', 'South Georgia and the South Sandwich Islands'),
    ('GT', 'Guatemala'),
    ('GU', 'Guam'),
    ('GW', 'Guinea-Bissau'),
    ('GY', 'Guyana'),
    ('HK', 'Hong Kong'),
    ('HM', 'Heard Island and McDonald Islands'),
    ('HN', 'Honduras'),
    ('HR', 'Croatia'),
    ('HT', 'Haiti'),
    ('HU', 'Hungary'),
    ('ID', 'Indonesia'),
    ('IE', 'Ireland'),
    ('IL', 'Israel'),
    ('IM', 'Isle of Man'),
    ('IN', 'India'),
    ('IO', 'British Indian Ocean Territory'),
    ('IQ', 'Iraq'),
    ('IR', 'Iran'),
    ('IS', 'Iceland'),
    ('IT', 'Italy'),
    ('JE', 'Jersey'),
    ('JM', 'Jamaica'),
    ('JO', 'Jordan'),
    ('JP', 'Japan'),
    ('KE', 'Kenya'),
    ('KG', 'Kyrgyzstan'),
    ('KH', 'Cambodia'),
    ('KI', 'Kiribati'),
    ('KM', 'Comoros'),
    ('KN', 'Saint Kitts and Nevis'),
    ('KP', 'Korea'),
    ('KR', 'Korea'),
    ('KW', 'Kuwait'),
    ('KY', 'Cayman Islands'),
    ('KZ', 'Kazakhstan'),
    ('LA', "Lao People's Democratic Republic"),
    ('LB', 'Lebanon'),
    ('LC', 'Saint Lucia'),
    ('LI', 'Liechtenstein'),
    ('LK', 'Sri Lanka'),
    ('LR', 'Liberia'),
    ('LS', 'Lesotho'),
    ('LT', 'Lithuania'),
    ('LU', 'Luxembourg'),
    ('LV', 'Latvia'),
    ('LY', 'Libyan Arab Jamahiriya'),
    ('MA', 'Morocco'),
    ('MC', 'Monaco'),
    ('MD', 'Moldova'),
    ('ME', 'Montenegro'),
    ('MG', 'Madagascar'),
    ('MH', 'Marshall Islands'),
    ('MK', 'Macedonia'),
    ('ML', 'Mali'),
    ('MM', 'Myanmar'),
    ('MN', 'Mongolia'),
    ('MO', 'Macao'),
    ('MP', 'Northern Mariana Islands'),
    ('MQ', 'Martinique'),
    ('MR', 'Mauritania'),
    ('MS', 'Montserrat'),
    ('MT', 'Malta'),
    ('MU', 'Mauritius'),
    ('MV', 'Maldives'),
    ('MW', 'Malawi'),
    ('MX', 'Mexico'),
    ('MY', 'Malaysia'),
    ('MZ', 'Mozambique'),
    ('NA', 'Namibia'),
    ('NC', 'New Caledonia'),
    ('NE', 'Niger'),
    ('NF', 'Norfolk Island'),
    ('NG', 'Nigeria'),
    ('NI', 'Nicaragua'),
    ('NL', 'Netherlands'),
    ('NO', 'Norway'),
    ('NP', 'Nepal'),
    ('NR', 'Nauru'),
    ('NU', 'Niue'),
    ('OM', 'Oman'),
    ('PA', 'Panama'),
    ('PE', 'Peru'),
    ('PF', 'French Polynesia'),
    ('PG', 'Papua New Guinea'),
    ('PH', 'Philippines'),
    ('PK', 'Pakistan'),
    ('PL', 'Poland'),
    ('PM', 'Saint Pierre and Miquelon'),
    ('PN', 'Pitcairn'),
    ('PR', 'Puerto Rico'),
    ('PS', 'Palestinian Territory'),
    ('PT', 'Portugal'),
    ('PW', 'Palau'),
    ('PY', 'Paraguay'),
    ('QA', 'Qatar'),
    ('RE', 'Reunion'),
    ('RO', 'Romania'),
    ('RS', 'Serbia'),
    ('RU', 'Russian Federation'),
    ('RW', 'Rwanda'),
    ('SA', 'Saudi Arabia'),
    ('SB', 'Solomon Islands'),
    ('SC', 'Seychelles'),
    ('SD', 'Sudan'),
    ('SE', 'Sweden'),
    ('SG', 'Singapore'),
    ('SH', 'Saint Helena'),
    ('SI', 'Slovenia'),
    ('SJ', 'Svalbard and Jan Mayen'),
    ('SK', 'Slovakia'),
    ('SL', 'Sierra Leone'),
    ('SM', 'San Marino'),
    ('SN', 'Senegal'),
    ('SO', 'Somalia'),
    ('SR', 'Suriname'),
    ('ST', 'Sao Tome and Principe'),
    ('SV', 'El Salvador'),
    ('SY', 'Syrian Arab Republic'),
    ('SZ', 'Swaziland'),
    ('TC', 'Turks and Caicos Islands'),
    ('TD', 'Chad'),
    ('TF', 'French Southern Territories'),
    ('TG', 'Togo'),
    ('TH', 'Thailand'),
    ('TJ', 'Tajikistan'),
    ('TK', 'Tokelau'),
    ('TL', 'Timor-Leste'),
    ('TM', 'Turkmenistan'),
    ('TN', 'Tunisia'),
    ('TO', 'Tonga'),
    ('TR', 'Turkey'),
    ('TT', 'Trinidad and Tobago'),
    ('TV', 'Tuvalu'),
    ('TW', 'Taiwan'),
    ('TZ', 'Tanzania'),
    ('UA', 'Ukraine'),
    ('UG', 'Uganda'),
    ('UM', 'United States Minor Outlying Islands'),
    ('UY', 'Uruguay'),
    ('UZ', 'Uzbekistan'),
    ('VA', 'Vatican City'),
    ('VC', 'Saint Vincent and the Grenadines'),
    ('VE', 'Venezuela'),
    ('VG', 'Virgin Islands (British)'),
    ('VI', 'Virgin Islands (US)'),
    ('VN', 'Viet Nam'),
    ('VU', 'Vanuatu'),
    ('WF', 'Wallis and Futuna'),
    ('WS', 'Samoa'),
    ('YE', 'Yemen'),
    ('YT', 'Mayotte'),
    ('ZA', 'South Africa'),
    ('ZM', 'Zambia'),
    ('ZW', 'Zimbabwe'),
)

ALPHABET = [
    'A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'J', 'K', 'L', 'M',
    'N', 'O', 'P', 'Q', 'R', 'S', 'T', 'U', 'V', 'W', 'X', 'Y', 'Z']



# Copyright 2010-2017, The University of Melbourne
# Copyright 2010-2017, Brian May
#
# This file is part of Karaage.
#
# Karaage is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Karaage is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Karaage  If not, see <http://www.gnu.org/licenses/>.

from django_xmlrpc.decorators import permission_required, xmlrpc_func

from karaage.common.decorators import xmlrpc_machine_required
from karaage.machines.models import Account
from karaage.projects.models import Project


@xmlrpc_machine_required()
@xmlrpc_func(returns='list', args=['string'])
def get_project_members(machine, project_id):
    """
    Returns list of usernames given a project id
    """
    try:
        project = Project.objects.get(pid=project_id)
    except Project.DoesNotExist:
        return 'Project not found'

    return [x.username for x in project.group.members.all()]


@xmlrpc_machine_required()
@xmlrpc_func(returns='list')
def get_projects(machine):
    """
    Returns list of project ids
    """

    query = Project.active.all()
    return [x.pid for x in query]


@xmlrpc_func(returns='string', args=['string', 'string', 'string'])
def get_project(username, project, machine_name=None):
    """
    Used in the submit filter to make sure user is in project
    """

    try:
        account = Account.objects.get(
            username=username,
            date_deleted__isnull=True)
    except Account.DoesNotExist:
        return "Account '%s' not found" % username

    if project is None:
        project = account.default_project
    else:
        try:
            project = Project.objects.get(pid=project)
        except Project.DoesNotExist:
            project = account.default_project

    if project is None:
        return "None"

    if account.person not in project.group.members.all():
        project = account.default_project

    if project is None:
        return "None"

    if account.person not in project.group.members.all():
        return "None"

    return project.pid


@permission_required()
@xmlrpc_func(returns='list')
def get_users_projects(user):
    """
    List projects a user is part of
    """
    person = user
    projects = person.projects.filter(is_active=True)
    return 0, [x.pid for x in projects]



# Copyright 2010-2017, The University of Melbourne
# Copyright 2010-2017, Brian May
#
# This file is part of Karaage.
#
# Karaage is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Karaage is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Karaage  If not, see <http://www.gnu.org/licenses/>.

import django_filters
import django_tables2 as tables
from django.db.models import Q
from django_tables2.utils import A

from karaage.projects.tables import ProjectColumn

from .models import CPUJob


class UnknownUsageFilter(django_filters.BooleanFilter):

    def filter(self, qs, value):
        if value is not None:
            if value:
                qs = qs.filter(
                    Q(project__isnull=True) | Q(account__isnull=True))
            else:
                qs = qs.filter(
                    Q(project__isnull=False) & Q(account__isnull=False))
        return qs


class CPUJobFilter(django_filters.FilterSet):
    person = django_filters.CharFilter(field_name="account__person__username")
    account = django_filters.CharFilter(field_name="account__pk")
    project = django_filters.CharFilter(field_name="project__pid")
    unknown_usage = UnknownUsageFilter()

    class Meta:
        model = CPUJob
        fields = ('person', 'queue', 'account', 'project')


class CPUJobTable(tables.Table):
    jobid = tables.LinkColumn('kg_usage_job_detail', args=[A('jobid')])
    project = ProjectColumn()
    person = tables.LinkColumn(
        'kg_person_detail', accessor="account.person",
        args=[A('account.person.username')])
    machine = tables.LinkColumn('kg_machine_detail', args=[A('machine.pk')])

    class Meta:
        model = CPUJob
        fields = ('jobid', 'person', 'project', 'machine', 'date',
                  'queue', 'cpu_usage', 'cores', 'vmem',
                  'wait_time', 'est_wall_time', 'act_wall_time')
        empty_text = "No items"



# -*- coding: utf-8 -*-
# Generated by Django 1.11.7 on 2018-01-31 17:58
from __future__ import unicode_literals

from django.conf import settings
from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('kgusage', '0002_auto_20170831_1735'),
        ('karaage', '0004_auto_20160429_0927'),
    ]

    run_before = [
        ('karaage', '0005_auto_20171215_1831'),
    ]

    operations = [
        migrations.AlterUniqueTogether(
            name='institutecache',
            unique_together=set([('date', 'start', 'end', 'institute')]),
        ),
        migrations.RemoveField(
            model_name='institutecache',
            name='machine_category',
        ),
        migrations.AlterUniqueTogether(
            name='machinecategorycache',
            unique_together=set([('date', 'start', 'end')]),
        ),
        migrations.RemoveField(
            model_name='machinecategorycache',
            name='machine_category',
        ),
        migrations.AlterUniqueTogether(
            name='personcache',
            unique_together=set([('date', 'start', 'end', 'person', 'project')]),
        ),
        migrations.RemoveField(
            model_name='personcache',
            name='machine_category',
        ),
        migrations.AlterUniqueTogether(
            name='projectcache',
            unique_together=set([('date', 'start', 'end', 'project')]),
        ),
        migrations.RemoveField(
            model_name='projectcache',
            name='machine_category',
        ),
    ]



# Copyright 2010-2017, The University of Melbourne
# Copyright 2010-2017, Brian May
#
# This file is part of Karaage.
#
# Karaage is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Karaage is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Karaage  If not, see <http://www.gnu.org/licenses/>.

import re

import six
from django.conf import settings

from karaage.datastores import account_exists
from karaage.machines.models import Account
from karaage.people.models import Person


username_re = re.compile(r'^%s$' % settings.USERNAME_VALIDATION_RE)


class UsernameException(Exception):
    pass


class UsernameInvalid(UsernameException):
    pass


class UsernameTaken(UsernameException):
    pass


def validate_username(username):
    """ Validate the new username. If the username is invalid, raises
    :py:exc:`UsernameInvalid`.

    :param username: Username to validate.
    """

    # Check username looks ok

    if not username.islower():
        raise UsernameInvalid(six.u('Username must be all lowercase'))
    if len(username) < 2:
        raise UsernameInvalid(six.u('Username must be at least 2 characters'))
    if not username_re.search(username):
        raise UsernameInvalid(settings.USERNAME_VALIDATION_ERROR_MSG)

    return username


def validate_username_for_new_person(username):
    """ Validate the new username for a new person. If the username is invalid
    or in use, raises :py:exc:`UsernameInvalid` or :py:exc:`UsernameTaken`.

    :param username: Username to validate.
    """

    # is the username valid?

    validate_username(username)

    # Check for existing people

    count = Person.objects.filter(username__exact=username).count()
    if count >= 1:
        raise UsernameTaken(six.u(
            'The username is already taken. Please choose another. '
            'If this was the name of your old account please email %s')
            % settings.ACCOUNTS_EMAIL)

    # Check for existing accounts

    count = Account.objects.filter(username__exact=username).count()
    if count >= 1:
        raise UsernameTaken(six.u(
            'The username is already taken. Please choose another. '
            'If this was the name of your old account please email %s')
            % settings.ACCOUNTS_EMAIL)

    # Check account datastore, in case username created outside Karaage.

    if account_exists(username):
        raise UsernameTaken(
            six.u('Username is already in external account datastore.'))

    return username


def validate_username_for_new_account(person, username):
    """ Validate the new username for a new account. If the username is invalid
    or in use, raises :py:exc:`UsernameInvalid` or :py:exc:`UsernameTaken`.

    :param person: Owner of new account.
    :param username: Username to validate.
    """

    # This is much the same as validate_username_for_new_person, except
    # we don't care if the username is used by the person owning the account

    # is the username valid?

    validate_username(username)

    # Check for existing people

    query = Person.objects.filter(username__exact=username)
    count = query.exclude(pk=person.pk).count()
    if count >= 1:
        raise UsernameTaken(six.u(
            'The username is already taken. Please choose another. '
            'If this was the name of your old account please email %s')
            % settings.ACCOUNTS_EMAIL)

    # Check for existing accounts not belonging to this person

    query = Account.objects.filter(username__exact=username)
    count = query.exclude(person__pk=person.pk).count()
    if count >= 1:
        raise UsernameTaken(six.u(
            'The username is already taken. Please choose another. '
            'If this was the name of your old account please email %s')
            % settings.ACCOUNTS_EMAIL)

    # Check datastore, in case username created outside Karaage.
    # Make sure we don't count the entry for person.

    query = Person.objects.filter(username__exact=username)
    count = query.filter(pk=person.pk).count()
    if count == 0 and account_exists(username):
        raise UsernameTaken(
            six.u('Username is already in external personal datastore.'))


def check_username_for_new_account(person, username):
    """ Check the new username for a new account. If the username  is
    in use, raises :py:exc:`UsernameTaken`.

    :param person: Owner of new account.
    :param username: Username to validate.
    """

    query = Account.objects.filter(
        username__exact=username,
        date_deleted__isnull=True)

    if query.count() > 0:
        raise UsernameTaken(
            six.u('Username already in use.')
        )

    if account_exists(username):
        raise UsernameTaken(
            six.u('Username is already in datastore.')
        )

    return username


def validate_username_for_rename_person(username, person):
    """ Validate the new username to rename a person. If the username is
    invalid or in use, raises :py:exc:`UsernameInvalid` or
    :py:exc:`UsernameTaken`.

    :param username: Username to validate.
    :param person: We exclude this person when checking for duplicates.
    """

    # is the username valid?

    validate_username(username)

    # Check for existing people

    count = Person.objects.filter(username__exact=username) \
        .exclude(pk=person.pk).count()
    if count >= 1:
        raise UsernameTaken(
            six.u('The username is already taken by an existing person.'))

    # Check for existing accounts not owned by person.
    # If there is a conflicting account owned by person it doesn't matter.

    count = Account.objects.filter(username__exact=username) \
        .exclude(person=person).count()
    if count >= 1:
        raise UsernameTaken(
            six.u('The username is already taken by an existing account.'))

    # Check datastore, in case username created outside Karaage.

    count = Account.objects.filter(
        username__exact=username,
        person=person,
        date_deleted__isnull=True).count()
    if count == 0 and account_exists(username):
        raise UsernameTaken(
            six.u('Username is already in external account datastore.'))



# Copyright 2010-2017, The University of Melbourne
# Copyright 2010-2017, Brian May
#
# This file is part of Karaage.
#
# Karaage is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Karaage is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Karaage  If not, see <http://www.gnu.org/licenses/>.

from django.conf.urls import url

from karaage.people.views import persons


urlpatterns = [
    url(r'^$', persons.user_detail, name='kg_person_detail'),
    url(r'^verbose/$', persons.user_verbose, name='kg_person_verbose'),
    url(r'^activate/$', persons.activate, name='kg_person_activate'),
    url(r'^delete/$', persons.delete_user, name='kg_person_delete'),
    url(r'^password/$', persons.password_change, name='kg_person_password'),
    url(r'^lock/$', persons.lock_person, name='kg_person_lock'),
    url(r'^unlock/$', persons.unlock_person, name='kg_person_unlock'),
    url(r'^bounced_email/$', persons.bounced_email, name='kg_person_bounce'),
    url(r'^logs/$', persons.person_logs, name='kg_person_logs'),
    url(r'^add_comment/$', persons.add_comment, name='kg_person_add_comment'),
    url(r'^edit/$', persons.edit_user, name='kg_person_edit'),
    url(r'^password_request/$',
        persons.password_request, name='kg_person_reset'),
    url(r'^password_request/done/$',
        persons.password_request_done, name='kg_person_reset_done'),
]



# Copyright 2010-2017, The University of Melbourne
# Copyright 2010-2017, Brian May
#
# This file is part of Karaage.
#
# Karaage is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Karaage is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Karaage  If not, see <http://www.gnu.org/licenses/>.
import os
import os.path

import environ


env = environ.Env()


class InvalidString(str):

    def __mod__(self, other):
        from django.template.base import TemplateSyntaxError
        raise TemplateSyntaxError(
            "Undefined variable or unknown value for: \"%s\"" % other)

# FIXME: TEMPLATES not accessible
# TEMPLATES[0]['OPTIONS']['string_if_invalid'] = InvalidString("%s")


DEBUG = True
SESSION_COOKIE_SECURE = False
PIPELINE_ENABLED = True
SHIB_SUPPORTED = True
SOUTH_TESTS_MIGRATE = False

LOGGING = {
    'version': 1,
    'disable_existing_loggers': True,
    'formatters': {
        'verbose': {
            'format': '%(levelname)s %(asctime)s '
            '%(module)s %(process)d %(thread)d %(message)s'
        },
        'simple': {
            'format': '%(levelname)s %(message)s'
        },
    },
    'handlers': {
        'null': {
            'level': 'DEBUG',
            'class': 'logging.NullHandler',
        },
        'console': {
            'level': 'DEBUG',
            'class': 'logging.StreamHandler',
            'formatter': 'verbose'
        },
    },
    'loggers': {
        '': {
            'handlers': ['console'],
            'level': 'WARNING',
        },
        'py.warnings': {
            # kg-manage test will automatically add a handler here
            # if verbosity is >= 0.
            'handlers': ["null"],
            'propagate': False,
        },
    },
}

# DATASTORES

DATASTORES = []

# OTHER

ACCOUNTS_ORG_NAME = 'TestOrg'

DATABASES = {'default': env.db(default="sqlite:///")}

if 'LDAP_URL' in os.environ:
    LDAP = {
        'default': {
            'ENGINE': 'tldap.backend.fake_transactions',
            'URI': os.environ['LDAP_URL'],
            'USER': os.environ['LDAP_DN'],
            'PASSWORD': os.environ['LDAP_PASSWORD'],
            'USE_TLS': False,  # Legacy, for TLDAP <= 0.2.16
            'REQUIRE_TLS': False,
            'START_TLS ': False,
            'TLS_CA': None,
        }
    }

SERVER_EMAIL = 'django@' + os.uname()[1]
ACCOUNTS_EMAIL = 'accounts@vpac.org'
EMAIL_SUBJECT_PREFIX = '[Grunt VPAC] - '
EMAIL_BACKEND = 'django.core.mail.backends.console.EmailBackend'

TIME_ZONE = 'Australia/Melbourne'
LANGUAGE_CODE = 'en-au'

TMP_DIR = 'tmp'

INTERNAL_IPS = (
    '127.0.0.1',
)

AUP_URL = 'http://example.com/aup.html'

REGISTRATION_BASE_URL = 'https://example.com/users'

SECRET_KEY = '5hvhpe6gv2t5x4$3dtq(w2v#vg@)sx4p3r_@wv%l41g!stslc*'

STATIC_ROOT = 'tmp/static'
STATIC_URL = "/static/"

TEST_MODULE_ROOT = os.path.dirname(os.path.realpath(__file__))
FIXTURE_DIRS = (os.path.join(TEST_MODULE_ROOT),)

ENABLE_CRACKLIB = False

MIN_PASSWORD_LENGTH = 6

PLUGINS = [
    'karaage.plugins.kgapplications.plugin',
    'karaage.plugins.kgsoftware.plugin',
    'karaage.plugins.kgsoftware.applications.plugin',
    'karaage.plugins.kgusage.plugin',
]

ALLOW_REGISTRATIONS = True



# Copyright 2010-2017, The University of Melbourne
# Copyright 2010-2017, Brian May
#
# This file is part of Karaage.
#
# Karaage is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Karaage is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Karaage  If not, see <http://www.gnu.org/licenses/>.

from __future__ import absolute_import

import logging

from django.conf import settings


LOG = logging.getLogger(__name__)


def assert_password_simple(password, old=None):
    if old and password == old:
        raise ValueError('Old and new passwords are the same.')
    elif len(password) < 6:
        raise ValueError('Password is less than six characters.')
    return password


try:
    from cracklib import VeryFascistCheck as _assert_password
    # Some configuration errors are only apparent when cracklib
    # tests a password for the first time, so test a strong password to
    # verify that cracklib is working as intended.
    _assert_password('thaeliez4niore0U')
except ImportError:
    _assert_password = assert_password_simple
except (OSError, ValueError) as e:
    LOG.warning("Cracklib misconfigured: %s", str(e))
    _assert_password = assert_password_simple


def assert_strong_password(username, password, old_password=None):
    """Raises ValueError if the password isn't strong.

    Returns the password otherwise."""

    # test the length
    try:
        minlength = settings.MIN_PASSWORD_LENGTH
    except AttributeError:
        minlength = 12
    if len(password) < minlength:
        raise ValueError(
            "Password must be at least %s characters long" % minlength)

    if username is not None and username in password:
        raise ValueError("Password contains username")

    return _assert_password(password, old_password)



# Copyright 2010-2017, The University of Melbourne
# Copyright 2010-2017, Brian May
#
# This file is part of Karaage.
#
# Karaage is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Karaage is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Karaage  If not, see <http://www.gnu.org/licenses/>.

from karaage.plugins import BasePlugin


class plugin(BasePlugin):
    name = "karaage.plugins.kgapplications"
    template_context_processors = (
        "karaage.plugins.kgapplications.context_processor",)

    settings = {
        'ALLOW_REGISTRATIONS': False,
        'ALLOW_NEW_PROJECTS': True,
        'APPROVE_ACCOUNTS_EMAIL': None,
    }

    def ready(self):
        from . import signals  # NOQA


def context_processor(request):
    """ Set context with common variables. """
    from .models import Application
    ctx = {}

    if request.user.is_authenticated:
        person = request.user
        my_applications = Application.objects.get_for_applicant(person)
        requires_attention = Application.objects.requires_attention(request)

        ctx['pending_applications'] = (
            my_applications.count() + requires_attention.count()
        )

    return ctx



import six
from django.contrib import messages
from django.contrib.auth.views import redirect_to_login
from django.core.exceptions import ImproperlyConfigured, ObjectDoesNotExist
from django.forms.models import ModelForm, ModelFormMetaclass
from django.http import Http404, HttpResponse, HttpResponseRedirect
from django.template import loader
from django.utils.translation import ugettext
from django.views.generic import GenericViewError


def apply_extra_context(extra_context, context):
    """
    Adds items from extra_context dict to context.  If a value in extra_context
    is callable, then it is called and the result is added to context.
    """
    for key, value in six.iteritems(extra_context):
        if callable(value):
            context[key] = value()
        else:
            context[key] = value


def get_model_and_form_class(model, form_class):
    """
    Returns a model and form class based on the model and form_class
    parameters that were passed to the generic view.

    If ``form_class`` is given then its associated model will be returned along
    with ``form_class`` itself.  Otherwise, if ``model`` is given, ``model``
    itself will be returned along with a ``ModelForm`` class created from
    ``model``.
    """
    if form_class:
        return form_class._meta.model, form_class
    if model:
        # The inner Meta class fails if model = model is used for some reason.
        tmp_model = model
        # TODO: we should be able to construct a ModelForm without creating
        # and passing in a temporary inner class.

        class Meta:
            model = tmp_model

        class_name = model.__name__ + 'Form'
        form_class = ModelFormMetaclass(
            class_name, (ModelForm,), {'Meta': Meta})
        return model, form_class
    raise GenericViewError("Generic view must be called with either a model or"
                           " form_class argument.")


def redirect(post_save_redirect, obj):
    """
    Returns a HttpResponseRedirect to ``post_save_redirect``.

    ``post_save_redirect`` should be a string, and can contain named string-
    substitution place holders of ``obj`` field names.

    If ``post_save_redirect`` is None, then redirect to ``obj``'s URL returned
    by ``get_absolute_url()``.  If ``obj`` has no ``get_absolute_url`` method,
    then raise ImproperlyConfigured.

    This function is meant to handle the post_save_redirect parameter to the
    ``create_object`` and ``update_object`` views.
    """
    if post_save_redirect:
        return HttpResponseRedirect(post_save_redirect % obj.__dict__)
    elif hasattr(obj, 'get_absolute_url'):
        return HttpResponseRedirect(obj.get_absolute_url())
    else:
        raise ImproperlyConfigured(
            "No URL to redirect to.  Either pass a post_save_redirect"
            " parameter to the generic view or define a get_absolute_url"
            " method on the Model.")


def lookup_object(model, object_id, slug, slug_field):
    """
    Return the ``model`` object with the passed ``object_id``.  If
    ``object_id`` is None, then return the object whose ``slug_field``
    equals the passed ``slug``.  If ``slug`` and ``slug_field`` are not passed,
    then raise Http404 exception.
    """
    lookup_kwargs = {}
    if object_id:
        lookup_kwargs['%s__exact' % model._meta.pk.name] = object_id
    elif slug and slug_field:
        lookup_kwargs['%s__exact' % slug_field] = slug
    else:
        raise GenericViewError(
            "Generic view must be called with either an object_id or a"
            " slug/slug_field.")
    try:
        return model.objects.get(**lookup_kwargs)
    except ObjectDoesNotExist:
        raise Http404("No %s found for %s"
                      % (model._meta.verbose_name, lookup_kwargs))


def create_object(
        request, model=None, template_name=None,
        template_loader=loader, extra_context=None, post_save_redirect=None,
        login_required=False, context_processors=None, form_class=None):
    """
    Generic object-creation function.

    Templates: ``<app_label>/<model_name>_form.html``
    Context:
        form
            the form for the object
    """
    if extra_context is None:
        extra_context = {}
    if login_required and not request.user.is_authenticated:
        return redirect_to_login(request.path)

    model, form_class = get_model_and_form_class(model, form_class)
    if request.method == 'POST':
        form = form_class(request.POST, request.FILES)
        if form.is_valid():
            new_object = form.save()

            msg = ugettext("The %(verbose_name)s was created successfully.") %\
                {"verbose_name": model._meta.verbose_name}
            messages.success(request, msg, fail_silently=True)
            return redirect(post_save_redirect, new_object)
    else:
        form = form_class()

    # Create the template, context, response
    if not template_name:
        template_name = "%s/%s_form.html" % (
            model._meta.app_label, model._meta.object_name.lower())
    t = template_loader.get_template(template_name)
    c = {
        'form': form,
    }
    apply_extra_context(extra_context, c)
    return HttpResponse(t.render(context=c, request=request))


def update_object(
        request, model=None, object_id=None, slug=None,
        slug_field='slug', template_name=None, template_loader=loader,
        extra_context=None, post_save_redirect=None, login_required=False,
        context_processors=None, template_object_name='object',
        form_class=None):
    """
    Generic object-update function.

    Templates: ``<app_label>/<model_name>_form.html``
    Context:
        form
            the form for the object
        object
            the original object being edited
    """
    if extra_context is None:
        extra_context = {}
    if login_required and not request.user.is_authenticated:
        return redirect_to_login(request.path)

    model, form_class = get_model_and_form_class(model, form_class)
    obj = lookup_object(model, object_id, slug, slug_field)

    if request.method == 'POST':
        form = form_class(request.POST, request.FILES, instance=obj)
        if form.is_valid():
            obj = form.save()
            msg = ugettext("The %(verbose_name)s was updated successfully.") %\
                {"verbose_name": model._meta.verbose_name}
            messages.success(request, msg, fail_silently=True)
            return redirect(post_save_redirect, obj)
    else:
        form = form_class(instance=obj)

    if not template_name:
        template_name = "%s/%s_form.html" % (
            model._meta.app_label, model._meta.object_name.lower())
    t = template_loader.get_template(template_name)
    c = {
        'form': form,
        template_object_name: obj,
    }
    apply_extra_context(extra_context, c)
    response = HttpResponse(t.render(context=c, request=request))
    return response


def delete_object(
        request, model, post_delete_redirect, object_id=None,
        slug=None, slug_field='slug', template_name=None,
        template_loader=loader, extra_context=None, login_required=False,
        context_processors=None, template_object_name='object'):
    """
    Generic object-delete function.

    The given template will be used to confirm deletetion if this view is
    fetched using GET; for safty, deletion will only be performed if this
    view is POSTed.

    Templates: ``<app_label>/<model_name>_confirm_delete.html``
    Context:
        object
            the original object being deleted
    """
    if extra_context is None:
        extra_context = {}
    if login_required and not request.user.is_authenticated:
        return redirect_to_login(request.path)

    obj = lookup_object(model, object_id, slug, slug_field)

    if request.method == 'POST':
        obj.delete()
        msg = ugettext("The %(verbose_name)s was deleted.") %\
            {"verbose_name": model._meta.verbose_name}
        messages.success(request, msg, fail_silently=True)
        return HttpResponseRedirect(post_delete_redirect)
    else:
        if not template_name:
            template_name = "%s/%s_confirm_delete.html" % (
                model._meta.app_label, model._meta.object_name.lower())
        t = template_loader.get_template(template_name)
        c = {
            template_object_name: obj,
        }
        apply_extra_context(extra_context, c)
        response = HttpResponse(t.render(context=c, request=request))
        return response



# Copyright 2010-2017, The University of Melbourne
# Copyright 2010-2017, Brian May
#
# This file is part of Karaage.
#
# Karaage is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Karaage is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Karaage  If not, see <http://www.gnu.org/licenses/>.

import os.path

from karaage.institutes.models import Institute

from .dirs import GRAPH_URL


try:
    import urlparse
except ImportError:
    import urllib.parse as urlparse


def get_colour(index):
    """ get color number index. """
    colours = [
        'red', 'blue', 'green', 'pink',
        'yellow', 'magenta', 'orange', 'cyan',
    ]
    default_colour = 'purple'
    if index < len(colours):
        return colours[index]
    else:
        return default_colour


# -----------------------------------------------------------------------


def get_project_trend_graph_filename(project, start, end):
    start_str = start.strftime('%Y-%m-%d')
    end_str = end.strftime('%Y-%m-%d')
    filename = os.path.join(
        "projects",
        "%s_%s_%s" % (project.pid, start_str, end_str)
    )
    return filename


def get_institute_graph_filename(start, end):
    start_str = start.strftime('%Y-%m-%d')
    end_str = end.strftime('%Y-%m-%d')
    filename = os.path.join(
        "institutes",
        "%s_%s" % (start_str, end_str)
    )
    return filename


def get_machine_graph_filename(start, end):
    start_str = start.strftime('%Y-%m-%d')
    end_str = end.strftime('%Y-%m-%d')
    filename = os.path.join(
        "machines",
        "%s_%s" % (start_str, end_str)
    )
    return filename


def get_trend_graph_filename(start, end):
    start_str = start.strftime('%Y-%m-%d')
    end_str = end.strftime('%Y-%m-%d')
    filename = os.path.join(
        "trends",
        "%s_%s" % (start_str, end_str)
    )
    return filename


def get_institute_trend_graph_filename(institute, start, end):
    start_str = start.strftime('%Y-%m-%d')
    end_str = end.strftime('%Y-%m-%d')
    filename = os.path.join(
        "i_trends",
        "%s_%s_%s" % (
            institute.name.replace(' ', '').replace('/', '-').lower(),
            start_str, end_str)
    )
    return filename


# -----------------------------------------------------------------------

def get_project_trend_graph_url(project, start, end):
    """Generates a bar graph for a project. """

    filename = get_project_trend_graph_filename(project, start, end)
    urls = {
        'graph_url': urlparse.urljoin(GRAPH_URL, filename + ".png"),
        'data_url': urlparse.urljoin(GRAPH_URL, filename + ".csv"),
    }

    return urls


def get_institute_graph_url(start, end):
    """ Pie chart comparing institutes usage. """

    filename = get_institute_graph_filename(start, end)
    urls = {
        'graph_url': urlparse.urljoin(GRAPH_URL, filename + ".png"),
        'data_url': urlparse.urljoin(GRAPH_URL, filename + ".csv"),
    }

    return urls


def get_machine_graph_url(start, end):
    """ Pie chart comparing machines usage. """

    filename = get_machine_graph_filename(start, end)
    urls = {
        'graph_url': urlparse.urljoin(GRAPH_URL, filename + ".png"),
        'data_url': urlparse.urljoin(GRAPH_URL, filename + ".csv"),
    }

    return urls


def get_trend_graph_url(start, end):
    """ Total trend graph for machine category. """

    filename = get_trend_graph_filename(start, end)
    urls = {
        'graph_url': urlparse.urljoin(GRAPH_URL, filename + ".png"),
        'data_url': urlparse.urljoin(GRAPH_URL, filename + ".csv"),
    }

    return urls


def get_institute_trend_graph_url(institute, start, end):
    """ Institute trend graph for machine category. """

    filename = get_institute_trend_graph_filename(institute, start, end)
    urls = {
        'graph_url': urlparse.urljoin(GRAPH_URL, filename + ".png"),
        'data_url': urlparse.urljoin(GRAPH_URL, filename + ".csv"),
    }

    return urls


def get_institutes_trend_graph_urls(start, end):
    """ Get all institute trend graphs. """

    graph_list = []
    for institute in Institute.objects.all():
        urls = get_institute_trend_graph_url(institute, start, end)
        urls['institute'] = institute
        graph_list.append(urls)

    return graph_list



import json

import six
from django.conf import settings
from django.contrib import messages
from django.db.models import Q
from django.http import (
    HttpResponse,
    HttpResponseBadRequest,
    HttpResponseRedirect,
)
from django.shortcuts import render

from karaage.common import log, saml
from karaage.institutes.models import Institute
from karaage.people.models import Person
from karaage.plugins.kgapplications import forms
from karaage.plugins.kgapplications.models import Applicant
from karaage.plugins.kgapplications.views import base
from karaage.plugins.kgapplications.views.states import StateWithSteps, Step
from karaage.projects.models import Project


def _get_applicant_from_saml(request):
    attrs, _ = saml.parse_attributes(request)
    saml_id = attrs['persistent_id']
    try:
        return Person.objects.get(saml_id=saml_id)
    except Person.DoesNotExist:
        pass

    try:
        return Applicant.objects.get(saml_id=saml_id)
    except Applicant.DoesNotExist:
        pass

    return None


class StateStepShibboleth(Step):

    """ Invitation has been sent to applicant. """
    name = "Invitation sent"

    def view(self, request, application, label, roles, actions):
        """ Django view method. """
        status = None
        applicant = application.applicant
        attrs = []

        saml_session = saml.is_saml_session(request)

        # certain actions are supported regardless of what else happens
        if 'cancel' in request.POST:
            return "cancel"
        if 'prev' in request.POST:
            return 'prev'

        # test for conditions where shibboleth registration not required
        if applicant.saml_id is not None:
            status = "You have already registered a shibboleth id."
            form = None
            done = True

        elif application.content_type.model != 'applicant':
            status = "You are already registered in the system."
            form = None
            done = True

        elif (applicant.institute is not None
                and applicant.institute.saml_entityid is None):
            status = "Your institute does not have shibboleth registered."
            form = None
            done = True

        elif Institute.objects.filter(
                saml_entityid__isnull=False).count() == 0:
            status = "No institutes support shibboleth here."
            form = None
            done = True

        else:
            # shibboleth registration is required

            # Do construct the form
            form = saml.SAMLInstituteForm(
                request.POST or None,
                initial={'institute': applicant.institute})
            done = False
            status = None

            # Was it a POST request?
            if request.method == 'POST':

                # Did the login form get posted?
                if 'login' in request.POST and form.is_valid():
                    institute = form.cleaned_data['institute']
                    applicant.institute = institute
                    applicant.save()
                    # We do not set application.insitute here, that happens
                    # when application, if it is a ProjectApplication, is
                    # submitted

                    # if institute supports shibboleth, redirect back here via
                    # shibboleth, otherwise redirect directly back he.
                    url = base.get_url(request, application, roles, label)
                    if institute.saml_entityid is not None:
                        url = saml.build_shib_url(
                            request, url, institute.saml_entityid)
                    return HttpResponseRedirect(url)

                # Did we get a logout request?
                elif 'logout' in request.POST:
                    if saml_session:
                        url = saml.logout_url(request)
                        return HttpResponseRedirect(url)
                    else:
                        return HttpResponseBadRequest("<h1>Bad Request</h1>")

            # did we get a shib session yet?
            if saml_session:
                attrs, _ = saml.parse_attributes(request)
                saml_session = True

        # if we are done, we can proceed to next state
        if request.method == 'POST':
            if 'cancel' in request.POST:
                return "cancel"
            if 'prev' in request.POST:
                return 'prev'

            if not done:
                if saml_session:
                    applicant = _get_applicant_from_saml(request)
                    if applicant is not None:
                        application.applicant = applicant
                        application.save()
                    else:
                        applicant = application.applicant

                    applicant = saml.add_saml_data(
                        applicant, request)
                    applicant.save()

                    done = True
                else:
                    status = "Please login to SAML before proceeding."

        if request.method == 'POST' and done:
            for action in actions:
                if action in request.POST:
                    return action
            return HttpResponseBadRequest("<h1>Bad Request</h1>")

        # render the page
        return render(
            template_name='kgapplications/project_aed_shibboleth.html',
            context={
                'form': form,
                'done': done,
                'status': status,
                'actions': actions,
                'roles': roles,
                'application': application,
                'attrs': attrs,
                'saml_session': saml_session,
            },
            request=request)


class StateStepApplicant(Step):

    """ Application is open and user is can edit it."""
    name = "Open"

    def view(self, request, application, label, roles, actions):
        """ Django view method. """
        # Get the appropriate form
        status = None
        form = None

        if application.content_type.model != 'applicant':
            status = "You are already registered in the system."
        elif application.content_type.model == 'applicant':
            if application.applicant.saml_id is not None:
                form = forms.SAMLApplicantForm(
                    request.POST or None,
                    instance=application.applicant)
            else:
                form = forms.UserApplicantForm(
                    request.POST or None,
                    instance=application.applicant)

        # Process the form, if there is one
        if form is not None and request.method == 'POST':
            if form.is_valid():
                form.save(commit=True)
                for action in actions:
                    if action in request.POST:
                        return action
                return HttpResponseBadRequest("<h1>Bad Request</h1>")
            else:
                # if form didn't validate and we want to go back or cancel,
                # then just do it.
                if 'cancel' in request.POST:
                    return "cancel"
                if 'prev' in request.POST:
                    return 'prev'

        # If we don't have a form, we can just process the actions here
        if form is None:
            for action in actions:
                if action in request.POST:
                    return action

        # Render the response
        return render(
            template_name='kgapplications/project_aed_applicant.html',
            context={
                'form': form,
                'application': application,
                'status': status, 'actions': actions, 'roles': roles,
            },
            request=request)


class StateStepProject(Step):

    """ Applicant is able to choose the project for the application. """
    name = "Choose project"

    def handle_ajax(self, request, application):
        resp = {}
        if 'leader' in request.POST:
            leader = Person.objects.get(pk=request.POST['leader'])
            project_list = leader.leads.filter(is_active=True)
            resp['project_list'] = \
                [(p.pk, six.text_type(p)) for p in project_list]

        elif 'terms' in request.POST:
            terms = request.POST['terms'].lower()
            try:
                project = Project.active.get(pid__icontains=terms)
                resp['project_list'] = [(project.pk, six.text_type(project))]
            except Project.DoesNotExist:
                resp['project_list'] = []
            except Project.MultipleObjectsReturned:
                resp['project_list'] = []
            leader_list = Person.active.filter(
                institute=application.applicant.institute,
                leads__is_active=True).distinct()
            if len(terms) >= 3:
                query = Q()
                for term in terms.split(' '):
                    q = Q(username__icontains=term)
                    q = q | Q(short_name__icontains=term)
                    q = q | Q(full_name__icontains=term)
                    query = query & q
                leader_list = leader_list.filter(query)
                resp['leader_list'] = [
                    (p.pk, "%s (%s)" % (p, p.username)) for p in leader_list]
            else:
                resp['error'] = "Please enter at lease three " \
                    "characters for search."
                resp['leader_list'] = []

        return resp

    def view(self, request, application, label, roles, actions):
        """ Django view method. """
        if 'ajax' in request.POST:
            resp = self.handle_ajax(request, application)
            return HttpResponse(
                json.dumps(resp), content_type="application/json")

        form_models = {
            'common': forms.CommonApplicationForm,
            'new': forms.NewProjectApplicationForm,
            'existing': forms.ExistingProjectApplicationForm,
        }

        project_forms = {}

        for key, form in six.iteritems(form_models):
            project_forms[key] = form(
                request.POST or None, instance=application)

        if application.project is not None:
            project_forms["common"].initial = {'application_type': 'U'}
        elif application.name != "":
            project_forms["common"].initial = {'application_type': 'P'}

        if 'application_type' in request.POST:
            at = request.POST['application_type']
            valid = True
            if at == 'U':
                # existing project
                if project_forms['common'].is_valid():
                    project_forms['common'].save(commit=False)
                else:
                    valid = False
                if project_forms['existing'].is_valid():
                    project_forms['existing'].save(commit=False)
                else:
                    valid = False

            elif at == 'P':
                # new project
                if project_forms['common'].is_valid():
                    project_forms['common'].save(commit=False)
                else:
                    valid = False
                if project_forms['new'].is_valid():
                    project_forms['new'].save(commit=False)
                else:
                    valid = False
                application.institute = application.applicant.institute

            else:
                return HttpResponseBadRequest("<h1>Bad Request</h1>")

            # reset hidden forms
            if at != 'U':
                # existing project form was not displayed
                project_forms["existing"] = (
                    form_models["existing"](instance=application))
                application.project = None
            if at != 'P':
                # new project form was not displayed
                project_forms["new"] = form_models["new"](instance=application)
                application.name = ""
                application.institute = None
                application.description = None
                application.additional_req = None
                application.machine_categories = []
                application.pid = None

            # save the values
            application.save()

            if project_forms['new'].is_valid() and at == 'P':
                project_forms["new"].save_m2m()

            # we still need to process cancel and prev even if form were
            # invalid
            if 'cancel' in request.POST:
                return "cancel"
            if 'prev' in request.POST:
                return 'prev'

            # if forms were valid, jump to next state
            if valid:
                for action in actions:
                    if action in request.POST:
                        return action
                return HttpResponseBadRequest("<h1>Bad Request</h1>")
        else:
            # we still need to process cancel, prev even if application type
            # not given
            if 'cancel' in request.POST:
                return "cancel"
            if 'prev' in request.POST:
                return 'prev'

        # lookup the project based on the form data
        project_id = project_forms['existing']['project'].value()
        project = None
        if project_id:
            try:
                project = Project.objects.get(pk=project_id)
            except Project.DoesNotExist:
                pass

        # render the response
        return render(
            template_name='kgapplications/project_aed_project.html',
            context={
                'forms': project_forms,
                'project': project,
                'actions': actions,
                'roles': roles,
                'application': application,
            },
            request=request)


class StateStepIntroduction(Step):

    """ Invitation has been sent to applicant. """
    name = "Read introduction"

    def view(self, request, application, label, roles, actions):
        """ Django get_next_action method. """
        if application.content_type.model == 'applicant':
            if not application.applicant.email_verified:
                application.applicant.email_verified = True
                application.applicant.save()
        for action in actions:
            if action in request.POST:
                return action
        link, is_secret = base.get_email_link(application)
        return render(
            template_name='kgapplications/project_aed_introduction.html',
            context={
                'actions': actions,
                'application': application, 'roles': roles,
                'link': link, 'is_secret': is_secret,
            },
            request=request)


class StateApplicantEnteringDetails(StateWithSteps):
    name = "Applicant entering details."
    actions = {'cancel', 'duplicate', 'reopen', 'submit'}

    def __init__(self, config):
        super(StateApplicantEnteringDetails, self).__init__(config)
        self.add_step(StateStepIntroduction(), 'intro')
        if settings.SHIB_SUPPORTED:
            self.add_step(StateStepShibboleth(), 'shibboleth')
        self.add_step(StateStepApplicant(), 'applicant')
        self.add_step(StateStepProject(), 'project')

    def get_actions(self, request, applications, roles):
        actions = set(self.actions)
        if 'is_applicant' not in roles:
            if 'submit' in actions:
                actions.remove('submit')
        return actions

    def get_next_action(self, request, application, label, roles):
        """ Process the get_next_action request at the current step. """

        # if user is logged and and not applicant, steal the
        # application
        if 'is_applicant' in roles:
            # if we got this far, then we either we are logged in as applicant,
            # or we know the secret for this application.

            new_person = None
            reason = None
            details = None

            attrs, _ = saml.parse_attributes(request)
            saml_id = attrs['persistent_id']
            if saml_id is not None:
                query = Person.objects.filter(saml_id=saml_id)
                if application.content_type.model == "person":
                    query = query.exclude(pk=application.applicant.pk)
                if query.count() > 0:
                    new_person = Person.objects.get(saml_id=saml_id)
                    reason = "SAML id is already in use by existing person."
                    details = (
                        "It is not possible to continue this application "
                        + "as is because the saml identity already exists "
                        + "as a registered user.")
                del query

            if request.user.is_authenticated:
                new_person = request.user
                reason = "%s was logged in " \
                    "and accessed the secret URL." % new_person
                details = (
                    "If you want to access this application "
                    + "as %s " % application.applicant
                    + "without %s stealing it, " % new_person
                    + "you will have to ensure %s is " % new_person
                    + "logged out first.")

            if new_person is not None:
                if application.applicant != new_person:
                    if 'steal' in request.POST:
                        old_applicant = application.applicant
                        application.applicant = new_person
                        application.save()
                        log.change(
                            application.application_ptr,
                            "Stolen application from %s" % old_applicant)
                        messages.success(
                            request,
                            "Stolen application from %s" % old_applicant)
                        url = base.get_url(request, application, roles, label)
                        return HttpResponseRedirect(url)
                    else:
                        return render(
                            template_name='kgapplications'
                            '/project_aed_steal.html',
                            context={
                                'application': application,
                                'person': new_person,
                                'reason': reason,
                                'details': details,
                            },
                            request=request)

        # if the user is the leader, show him the leader specific page.
        if ('is_leader' in roles or 'is_delegate' in roles) \
                and 'is_admin' not in roles \
                and 'is_applicant' not in roles:
            actions = ['reopen']
            if 'reopen' in request.POST:
                return 'reopen'
            return render(
                template_name='kgapplications/project_aed_for_leader.html',
                context={'application': application,
                         'actions': actions, 'roles': roles, },
                request=request)

        # otherwise do the default behaviour for StateWithSteps
        return super(StateApplicantEnteringDetails, self) \
            .get_next_action(request, application, label, roles)



# Copyright 2010-2017, The University of Melbourne
# Copyright 2010-2017, Brian May
#
# This file is part of Karaage.
#
# Karaage is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Karaage is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Karaage  If not, see <http://www.gnu.org/licenses/>.

from karaage.plugins import BasePlugin


class plugin(BasePlugin):
    name = "karaage.plugins.kgusage"
    xmlrpc_methods = (
        ('karaage.plugins.kgusage.xmlrpc.parse_usage',
            'parse_usage'),
        ('karaage.plugins.kgusage.xmlrpc.add_modules_used',
            'add_modules_used'),
    )
    settings = {
        'GRAPH_DEBUG': False,
        'GRAPH_DIR': 'kgusage/',
        'GRAPH_TMP': 'kgusage/',
    }
    depends = ("karaage.plugins.kgsoftware.plugin",)

    def ready(self):
        from . import signals  # NOQA



# Copyright 2010-2017, The University of Melbourne
# Copyright 2010-2017, Brian May
#
# This file is part of Karaage.
#
# Karaage is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Karaage is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Karaage  If not, see <http://www.gnu.org/licenses/>.

import datetime

from django.conf import settings
from django.contrib.auth.models import AbstractBaseUser
from django.db import models
from django.urls import reverse
from django.utils.encoding import python_2_unicode_compatible
from jsonfield import JSONField
from model_utils import FieldTracker

from karaage.common import is_admin, log
from karaage.machines.managers import ActiveMachineManager, MachineManager
from karaage.people.models import Group, Person


@python_2_unicode_compatible
class Machine(AbstractBaseUser):
    name = models.CharField(max_length=50, unique=True)
    no_cpus = models.IntegerField()
    no_nodes = models.IntegerField()
    type = models.CharField(max_length=100)
    start_date = models.DateField()
    end_date = models.DateField(null=True, blank=True)
    pbs_server_host = models.CharField(max_length=50, null=True, blank=True)
    mem_per_core = models.IntegerField(
        help_text="In GB", null=True, blank=True)
    objects = MachineManager()
    active = ActiveMachineManager()
    scaling_factor = models.IntegerField(default=1)

    USERNAME_FIELD = 'name'

    _tracker = FieldTracker()

    def save(self, *args, **kwargs):
        created = self.pk is None

        # save the object
        super(Machine, self).save(*args, **kwargs)

        if created:
            log.add(self, 'Created')
        for field in self._tracker.changed():
            if field == "password":
                log.change(self, 'Changed %s' % field)
            else:
                log.change(
                    self,
                    'Changed %s to %s' % (field, getattr(self, field)))

    def delete(self, *args, **kwargs):
        # delete the object
        log.delete(self, 'Deleted')
        super(Machine, self).delete(*args, **kwargs)

    class Meta:
        db_table = 'machine'
        app_label = 'karaage'

    def __str__(self):
        return self.name

    def get_absolute_url(self):
        return reverse('kg_machine_detail', args=[self.id])


@python_2_unicode_compatible
class Account(models.Model):
    person = models.ForeignKey(Person, on_delete=models.CASCADE)
    username = models.CharField(max_length=255)
    foreign_id = models.CharField(
        max_length=255, null=True, unique=True,
        help_text='The foreign identifier from the datastore.')
    default_project = models.ForeignKey(
        'karaage.Project', null=True, blank=True,
        on_delete=models.SET_NULL)
    date_created = models.DateField()
    date_deleted = models.DateField(null=True, blank=True)
    disk_quota = models.IntegerField(null=True, blank=True, help_text="In GB")
    shell = models.CharField(max_length=50)
    login_enabled = models.BooleanField(default=True)
    extra_data = JSONField(
        default={},
        help_text='Datastore specific values should be stored in this field.')

    _tracker = FieldTracker()

    def __init__(self, *args, **kwargs):
        super(Account, self).__init__(*args, **kwargs)
        self._password = None

    class Meta:
        ordering = ['person', ]
        db_table = 'account'
        app_label = 'karaage'

    def __str__(self):
        return '%s' % self.username

    def get_absolute_url(self):
        return reverse('kg_account_detail', args=[self.pk])

    @classmethod
    def create(cls, person, default_project):
        """Creates a Account (if needed) and activates person.
        """
        ua = Account.objects.create(
            person=person, username=person.username,
            shell=settings.DEFAULT_SHELL,
            default_project=default_project,
            date_created=datetime.datetime.today())

        if default_project is not None:
            person.add_group(default_project.group)

        return ua

    def project_list(self):
        return self.person.projects.all()

    def save(self, *args, **kwargs):
        created = self.pk is None

        # save the object
        super(Account, self).save(*args, **kwargs)

        if created:
            log.add(
                self.person,
                'Account %s: Created' % self)
        for field in self._tracker.changed():
            if field != "password":
                log.change(
                    self.person,
                    'Account %s: Changed %s to %s'
                    % (self, field, getattr(self, field)))

        # check if it was renamed
        if self._tracker.has_changed('username'):
            old_username = self._tracker.previous('username')
            if old_username is not None:
                new_username = self.username
                if self.date_deleted is None:
                    from karaage.datastores import set_account_username
                    set_account_username(self, old_username, new_username)
                log.change(
                    self.person,
                    'Account %s: Changed username from %s to %s' %
                    (self, old_username, new_username))

        # check if deleted status changed
        if self._tracker.has_changed('date_deleted'):
            if self.date_deleted is not None:
                # account is deactivated
                from karaage.datastores import delete_account
                delete_account(self)
                log.delete(
                    self.person,
                    'Account %s: Deactivated account' % self)
                # deleted
            else:
                # account is reactivated
                log.add(
                    self.person,
                    'Account %s: Activated' % self)

        # makes sense to lock non-existant account
        if self.date_deleted is not None:
            self.login_enabled = False

        # update the datastore
        if self.date_deleted is None:
            from karaage.datastores import save_account
            save_account(self)

            if self._password is not None:
                from karaage.datastores import set_account_password
                set_account_password(self, self._password)
                log.change(
                    self.person,
                    'Account %s: Changed Password' % self)
                self._password = None
    save.alters_data = True

    def can_view(self, request):
        # if user not authenticated, no access
        if not request.user.is_authenticated:
            return False

        # ensure person making request isn't deleted.
        if not request.user.is_active:
            return False

        # ensure person making request isn't locked.
        if request.user.is_locked():
            return False

        # if user is admin, full access
        if is_admin(request):
            return True

        # ensure this account is not locked
        if self.is_locked():
            return False

        # ensure this account is not deleted
        if self.date_deleted is not None:
            return False

        # ensure person owning account isn't locked.
        if self.person.is_locked():
            return False

        # ensure person owning account isn't deleted.
        if not self.person.is_active:
            return False

        return True

    def can_edit(self, request):
        # if we can't view this account, we can't edit it either
        if not self.can_view(request):
            return False

        if not is_admin(request):
            # if not admin, ensure we are the person being altered
            if self.person != request.user:
                return False

        return True

    def delete(self, **kwargs):
        # delete the object
        log.delete(self.person, 'Account %s: Deleted' % self)
        super(Account, self).delete(**kwargs)
        if self.date_deleted is None:
            # delete the datastore
            from karaage.datastores import delete_account
            delete_account(self)
    delete.alters_data = True

    def deactivate(self):
        if self.date_deleted is not None:
            raise RuntimeError("Account is deactivated")
        # save the object
        self.date_deleted = datetime.datetime.now()
        self.login_enabled = False
        self.save()
        # self.save() will delete the datastore for us.
    deactivate.alters_data = True

    def change_shell(self, shell):
        self.shell = shell
        self.save()
        # self.save() will update the datastore for us.
    change_shell.alters_data = True

    def set_password(self, password):
        if self.date_deleted is not None:
            raise RuntimeError("Account is deactivated")
        self._password = password
    set_password.alters_data = True

    def get_disk_quota(self):
        # FIXME: should this become deprecated?
        return self.disk_quota

    def login_shell(self):
        return self.shell

    def lock(self):
        if self.date_deleted is not None:
            raise RuntimeError("Account is deactivated")
        self.login_enabled = False
        self.save()
    lock.alters_data = True

    def unlock(self):
        if self.date_deleted is not None:
            raise RuntimeError("Account is deactivated")
        self.login_enabled = True
        self.save()
    unlock.alters_data = True

    def is_locked(self):
        return not self.login_enabled


def _remove_group(group, person):
    # if removing default project from person, then break link first
    for ua in person.account_set.filter(
            date_deleted__isnull=True, default_project__isnull=False):
        # Does the default_project for ua belong to this group?
        count = group.project_set.filter(pk=ua.default_project.pk).count()
        # If yes, deactivate the ua
        if count > 0:
            ua.default_project = None
            ua.save()


def _members_changed(
        sender, instance, action, reverse, model, pk_set, **kwargs):
    """
    Hook that executes whenever the group members are changed.
    """
    if action == "post_remove":
        if not reverse:
            group = instance
            for person in model.objects.filter(pk__in=pk_set):
                _remove_group(group, person)
        else:
            person = instance
            for group in model.objects.filter(pk__in=pk_set):
                _remove_group(group, person)

    elif action == "pre_clear":
        # This has to occur in pre_clear, not post_clear, as otherwise
        # we won't see what groups need to be removed.
        if not reverse:
            group = instance
            for person in group.members.all():
                _remove_group(group, person)
        else:
            person = instance
            for group in person.groups.all():
                _remove_group(group, person)


models.signals.m2m_changed.connect(
    _members_changed, sender=Group.members.through)



# Copyright 2010-2017, The University of Melbourne
# Copyright 2010-2017, Brian May
#
# This file is part of Karaage.
#
# Karaage is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Karaage is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Karaage  If not, see <http://www.gnu.org/licenses/>.

import re
import sys

import django.db.transaction
import tldap.transaction
from django.conf import settings
from django.core.management.base import BaseCommand, CommandError

from karaage.projects.models import Project


try:
    input = raw_input
except NameError:
    pass


class Command(BaseCommand):
    help = 'Change a pid for a project and all accounts for that project'

    def add_arguments(self, parser):
        parser.add_argument('old_pid', type=str)
        parser.add_argument('new_pid', type=str)

    @django.db.transaction.atomic
    @tldap.transaction.commit_on_success
    def handle(self, *args, **options):
        old = options['old_pid']
        new = options['new_pid']

        try:
            project = Project.objects.get(pid=old)
        except Project.DoesNotExist:
            raise CommandError('project %s does not exist' % old)

        project_re = re.compile(r'^%s$' % settings.PROJECT_VALIDATION_RE)
        if not project_re.search(new):
            raise CommandError(settings.PROJECT_VALIDATION_ERROR_MSG)

        while True:
            confirm = input(
                'Change project "%s" to "%s (yes,no): ' % (old, new))
            if confirm == 'yes':
                break
            elif confirm == 'no':
                return sys.exit(0)
            else:
                print("Please enter yes or no")

        project.pid = new
        project.save()
        print("Changed pid on project")

        print("Done")



# Copyright 2010-2017, The University of Melbourne
# Copyright 2010-2017, Brian May
#
# This file is part of Karaage.
#
# Karaage is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Karaage is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Karaage  If not, see <http://www.gnu.org/licenses/>.

from django.conf import settings
from django.contrib import messages
from django.core.mail import send_mass_mail
from django.http import HttpResponseRedirect
from django.shortcuts import render
from django.template import Context, Template
from django.urls import reverse

from karaage.common.decorators import admin_required
from karaage.emails.forms import BulkEmailForm


def _get_emails(person_query, subject, body):
    emails = []

    if person_query:
        email_list = []

        subject_t = Template(subject)
        body_t = Template(body)

        person_query = person_query.filter(
            is_systemuser=False, login_enabled=True)

        for person in person_query:
            if person.email not in email_list:
                projects = ", ".join(
                    [str(project) for project in person.leads.all()]
                )
                ctx = Context({
                    'projects': projects,
                    'receiver': person,
                })
                subject = subject_t.render(ctx)
                body = body_t.render(ctx)
                emails.append(
                    (subject, body, settings.ACCOUNTS_EMAIL,
                        [person.email])
                )
                email_list.append(person.email)

    return emails


@admin_required
def send_email(request):

    form = BulkEmailForm(request.POST or None)
    if request.method == 'POST':

        if form.is_valid():
            subject = form.cleaned_data['subject']
            body = form.cleaned_data['body']
            person_query = form.get_person_query()
            emails = _get_emails(person_query, subject, body)

            if 'preview' in request.POST:
                try:
                    preview = emails[0]
                except IndexError:
                    preview = None
            else:
                send_mass_mail(emails)
                messages.success(request, "Emails sent successfully")

                return HttpResponseRedirect(reverse('index'))

    return render(
        template_name='karaage/emails/send_email_form.html', context=locals(),
        request=request)



from django.contrib import messages

from karaage.plugins.kgapplications import emails
from karaage.plugins.kgapplications.views import base


class TransitionOpen(base.Transition):
    """ A transition after application opened. """
    actions = {'success'}

    def get_next_action(self, request, application, roles):
        """ Retrieve the next state. """
        application.reopen()
        link, is_secret = base.get_email_link(application)
        emails.send_invite_email(application, link, is_secret)
        messages.success(
            request,
            "Sent an invitation to %s."
            % application.applicant.email)
        return 'success'


class TransitionSubmit(base.Transition):
    """ A transition after application submitted. """
    actions = {'success', 'error'}

    def get_next_action(self, request, application, roles):
        """ Retrieve the next state. """

        # Check for serious errors in submission.
        # Should only happen in rare circumstances.
        errors = application.check_valid()
        if len(errors) > 0:
            for error in errors:
                messages.error(request, error)
            return 'error'

        # mark as submitted
        application.submit()

        return 'success'


class TransitionApprove(base.Transition):
    """ A transition after application fully approved. """
    actions = {'password_needed', 'password_ok', 'error'}

    def get_next_action(self, request, application, roles):
        """ Retrieve the next state. """
        # Check for serious errors in submission.
        # Should only happen in rare circumstances.
        errors = application.check_valid()
        if len(errors) > 0:
            for error in errors:
                messages.error(request, error)
            return 'error'

        # approve application
        approved_by = request.user
        created_person, created_account = application.approve(approved_by)

        # send email
        application.extend()
        link, is_secret = base.get_email_link(application)
        emails.send_approved_email(
            application, created_person, created_account, link, is_secret)

        if created_person or created_account:
            return 'password_needed'
        else:
            return 'password_ok'


class TransitionSplit(base.Transition):
    """ A transition after application submitted. """
    actions = {'existing_project', 'new_project', 'error'}

    def get_next_action(self, request, application, roles):
        """ Retrieve the next state. """
        # Do we need to wait for leader or delegate approval?
        if application.project is None:
            return 'new_project'
        else:
            return 'existing_project'



# Copyright 2010-2017, The University of Melbourne
# Copyright 2010-2017, Brian May
#
# This file is part of Karaage.
#
# Karaage is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Karaage is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Karaage  If not, see <http://www.gnu.org/licenses/>.

import django_tables2 as tables
import six
from django.contrib import messages
from django.http import HttpResponseRedirect
from django.shortcuts import get_object_or_404, render
from django.urls import reverse

import karaage.common as util
from karaage.common.decorators import admin_required
from karaage.people.forms import AddGroupMemberForm, AdminGroupForm
from karaage.people.models import Group, Person
from karaage.people.tables import GroupFilter, GroupTable


@admin_required
def group_list(request):
    queryset = Group.objects.select_related()

    q_filter = GroupFilter(request.GET, queryset=queryset)
    table = GroupTable(q_filter.qs)
    tables.RequestConfig(request).configure(table)

    spec = []
    for name, value in six.iteritems(q_filter.form.cleaned_data):
        if value is not None and value != "":
            name = name.replace('_', ' ').capitalize()
            spec.append((name, value))

    return render(
        template_name='karaage/people/group_list.html',
        context={
            'table': table,
            'filter': q_filter,
            'spec': spec,
            'title': "Group list",
        },
        request=request)


def _add_edit_group(request, form_class, group_name):
    group_form = form_class

    if group_name is None:
        group = None
    else:
        group = get_object_or_404(Group, name=group_name)

    if request.method == 'POST':
        form = group_form(request.POST, instance=group)
        if form.is_valid():
            if group:
                # edit
                group = form.save()
                messages.success(
                    request, "Group '%s' was edited succesfully" % group)
            else:
                # add
                group = form.save()
                messages.success(
                    request, "Group '%s' was created succesfully" % group)

            return HttpResponseRedirect(group.get_absolute_url())
    else:
        form = group_form(instance=group)

    return render(
        template_name='karaage/people/group_form.html',
        context={'group': group, 'form': form},
        request=request)


@admin_required
def add_group(request):
    return _add_edit_group(request, AdminGroupForm, None)


@admin_required
def edit_group(request, group_name):
    return _add_edit_group(request, AdminGroupForm, group_name)


@admin_required
def delete_group(request, group_name):
    group = get_object_or_404(Group, name=group_name)

    error = None

    if group.institute_set.all().count() > 0:
        error = "An institute is using this group."
    elif group.project_set.all().count() > 0:
        error = "A project is using this group."
    elif request.method == 'POST':
        group.delete()
        messages.success(request, "Group '%s' was deleted succesfully" % group)
        return HttpResponseRedirect(reverse("kg_group_list"))

    return render(
        template_name='karaage/people/group_confirm_delete.html',
        context=locals(),
        request=request)


@admin_required
def group_detail(request, group_name):
    group = get_object_or_404(Group, name=group_name)
    form = AddGroupMemberForm(instance=group)

    return render(
        template_name='karaage/people/group_detail.html',
        context=locals(),
        request=request)


@admin_required
def group_verbose(request, group_name):
    group = get_object_or_404(Group, name=group_name)

    from karaage.datastores import get_group_details
    group_details = get_group_details(group)

    return render(
        template_name='karaage/people/group_verbose.html',
        context=locals(),
        request=request)


@admin_required
def group_logs(request, group_name):
    obj = get_object_or_404(Group, name=group_name)
    breadcrumbs = [
        ("Groups", reverse("kg_group_list")),
        (six.text_type(obj), reverse("kg_group_detail", args=[obj.name]))
    ]
    return util.log_list(request, breadcrumbs, obj)


@admin_required
def add_comment(request, group_name):
    obj = get_object_or_404(Group, name=group_name)
    breadcrumbs = [
        ("Groups", reverse("kg_group_list")),
        (six.text_type(obj), reverse("kg_group_detail", args=[obj.name]))
    ]
    return util.add_comment(request, breadcrumbs, obj)


@admin_required
def add_group_member(request, group_name):
    group = get_object_or_404(Group, name=group_name)
    if request.method == 'POST':
        form = AddGroupMemberForm(request.POST, instance=group)
        if form.is_valid():
            form.save()
            return HttpResponseRedirect(group.get_absolute_url())
    else:
        form = AddGroupMemberForm(instance=group)

    return render(
        template_name='karaage/people/group_add_member.html',
        context=locals(),
        request=request)


@admin_required
def remove_group_member(request, group_name, username):
    group = get_object_or_404(Group, name=group_name)
    person = get_object_or_404(Person, username=username)

    error = None
    for account in person.account_set.filter(date_deleted__isnull=True):
        # Does the default_project for ua belong to this group?
        count = group.project_set.filter(pk=account.default_project.pk).count()
        # If yes, error
        if count > 0:
            error = "The person has accounts that use " \
                    "this group as the default_project."

    if error is None and request.method == 'POST':
        group.remove_person(person)
        return HttpResponseRedirect(group.get_absolute_url())

    return render(
        template_name='karaage/people/group_remove_member.html',
        context=locals(),
        request=request)



# Copyright 2010-2017, The University of Melbourne
# Copyright 2010-2017, Brian May
#
# This file is part of Karaage.
#
# Karaage is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Karaage is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Karaage  If not, see <http://www.gnu.org/licenses/>.

""" Slurm datastore. """

import csv
import logging
import subprocess
import sys

import karaage.common.trace as trace
from karaage.datastores import base


logger = logging.getLogger(__name__)

if sys.version_info < (3, 0):
    # Python2: csv module does not support unicode, we must use byte strings.

    def _input_csv(csv_data):
        for line in csv_data:
            assert isinstance(line, bytes)
            yield line

    def _output_csv(csv_line):
        for i, column in enumerate(csv_line):
            csv_line[i] = column.decode("ascii", errors='ignore')
            assert isinstance(csv_line[i], unicode)  # NOQA

else:
    # Python3: csv module does support unicode, we must use strings everywhere,
    # not byte strings

    def _input_csv(unicode_csv_data):
        for line in unicode_csv_data:
            assert isinstance(line, bytes)
            line = line.decode("ascii", errors='ignore')
            assert isinstance(line, str)
            yield line

    def _output_csv(csv_line):
        for column in csv_line:
            assert isinstance(column, str)


class SlurmDataStore(base.DataStore):
    """ Slurm datastore. """

    def __init__(self, config):
        super(SlurmDataStore, self).__init__(config)
        self._prefix = config.get('PREFIX', ["sudo", "-uslurm"])
        self._path = config.get('PATH', "/usr/local/slurm/latest/bin/sacctmgr")
        self._null_project = config.get('NULL_PROJECT', "default")
        self._add_account_extra = config.get('ADD_ACCOUNT_EXTRA', ['grpcpumins=0'])

    @staticmethod
    def _filter_string(value):
        """ Filter the string so Gold doesn't have heart failure."""
        if value is None:
            value = ""

        # replace whitespace with space
        value = value.replace("\n", " ")
        value = value.replace("\t", " ")

        # CSV seperator
        value = value.replace("|", " ")

        # remove leading/trailing whitespace
        value = value.strip()

        # Used for stripping non-ascii characters
        value = ''.join(c for c in value if 31 < ord(c) < 127)

        return value

    @staticmethod
    def _truncate(value, arg):
        """
        Truncates a string after a given number of chars
        Argument: Number of chars to _truncate after
        """
        length = int(arg)
        if value is None:
            value = ""
        if len(value) > length:
            return value[:length] + "..."
        else:
            return value

    def _call(self, command, ignore_errors=None):
        """ Call remote command with logging. """
        if ignore_errors is None:
            ignore_errors = []
        cmd = []
        cmd.extend(self._prefix)
        cmd.extend([self._path, "-iP"])
        cmd.extend(command)
        command = cmd

        logger.debug("Cmd %s" % command)
        null = open('/dev/null', 'w')
        retcode = subprocess.call(command, stdout=null, stderr=null)
        null.close()

        if retcode in ignore_errors:
            logger.debug(
                "<-- Cmd %s returned %d (ignored)" % (command, retcode))
            return

        if retcode:
            logger.error(
                "<-- Cmd %s returned: %d (error)" % (command, retcode))
            raise subprocess.CalledProcessError(retcode, command)

        logger.debug("<-- Returned %d (good)" % retcode)
        return

    def _read_output(self, command):
        """ Read CSV delimited input from Slurm. """
        cmd = []
        cmd.extend(self._prefix)
        cmd.extend([self._path, "-iP"])
        cmd.extend(command)
        command = cmd

        logger.debug("Cmd %s" % command)
        null = open('/dev/null', 'w')
        process = subprocess.Popen(
            command, stdout=subprocess.PIPE, stderr=null)
        null.close()

        results = []
        reader = csv.reader(_input_csv(process.stdout), delimiter=str("|"))

        try:
            headers = next(reader)
            logger.debug("<-- headers %s" % headers)
        except StopIteration:
            logger.debug("Cmd %s headers not found" % command)
            headers = []

        for row in reader:
            _output_csv(row)

            logger.debug("<-- row %s" % row)
            this_row = {}

            i = 0
            for i in range(0, len(headers)):
                key = headers[i]
                value = row[i]
                this_row[key] = value

            results.append(this_row)

        process.stdout.close()
        retcode = process.wait()
        if retcode != 0:
            logger.error("<-- Cmd %s returned %d (error)" % (command, retcode))
            raise subprocess.CalledProcessError(retcode, command)

        if len(headers) == 0:
            logger.error("Cmd %s didn't return any headers." % command)
            raise RuntimeError("Cmd %s didn't return any headers." % command)

        logger.debug("<-- Returned: %d (good)" % retcode)
        return results

    def get_user(self, username):
        """ Get the user details from Slurm. """
        cmd = ["list", "user", "where", "name=%s" % username]
        results = self._read_output(cmd)

        if len(results) == 0:
            return None
        elif len(results) > 1:
            logger.error(
                "Command returned multiple results for '%s'." % username)
            raise RuntimeError(
                "Command returned multiple results for '%s'." % username)

        the_result = results[0]
        the_name = the_result["User"]
        if username.lower() != the_name.lower():
            logger.error(
                "We expected username '%s' but got username '%s'."
                % (username, the_name))
            raise RuntimeError(
                "We expected username '%s' but got username '%s'."
                % (username, the_name))

        return the_result

    def get_project(self, projectname):
        """ Get the project details from Slurm. """
        cmd = ["list", "accounts", "where", "name=%s" % projectname]
        results = self._read_output(cmd)

        if len(results) == 0:
            return None
        elif len(results) > 1:
            logger.error(
                "Command returned multiple results for '%s'." % projectname)
            raise RuntimeError(
                "Command returned multiple results for '%s'." % projectname)

        the_result = results[0]
        the_project = the_result["Account"]
        if projectname.lower() != the_project.lower():
            logger.error(
                "We expected projectname '%s' "
                "but got projectname '%s'." % (projectname, the_project))
            raise RuntimeError(
                "We expected projectname '%s' "
                "but got projectname '%s'." % (projectname, the_project))

        return the_result

    def get_users_in_project(self, projectname):
        """ Get list of users in project from Slurm. """
        cmd = ["list", "assoc", "where", "account=%s" % projectname]
        results = self._read_output(cmd)

        user_list = []
        for result in results:
            if result["User"] != "":
                user_list.append(result["User"])
        return user_list

    def get_projects_in_user(self, username):
        """ Get list of projects in user from Slurm. """
        cmd = ["list", "assoc", "where", "user=%s" % username]
        results = self._read_output(cmd)

        project_list = []
        for result in results:
            project_list.append(result["Account"])
        return project_list

    def is_user_in_project(self, username, projectname):
        cmd = ["list", "assoc", "where",
               "user=%s" % username, "account=%s" % projectname]
        results = self._read_output(cmd)
        return len(results) > 0

    def _save_account(self, account, username):
        """ Called when account is created/updated. With username override. """

        # retrieve default project, or use null project if none
        default_project_name = self._null_project
        if account.default_project is not None:
            default_project_name = account.default_project.pid

        # account created
        # account updated

        ds_user = self.get_user(username)
        if account.date_deleted is None:
            # date_deleted is not set, user should exist
            logger.debug("account is active")

            if ds_user is None:
                # create user if doesn't exist
                self._call([
                    "add", "user",
                    "accounts=%s" % default_project_name,
                    "defaultaccount=%s" % default_project_name,
                    "name=%s" % username])
            else:
                # or just set default project
                self._call([
                    "modify", "user",
                    "set", "defaultaccount=%s" % default_project_name,
                    "where", "name=%s" % username])

            # update user meta information

            # add rest of projects user belongs to
            slurm_projects = self.get_projects_in_user(username)
            slurm_projects = [project.lower() for project in slurm_projects]
            slurm_projects = set(slurm_projects)
            for project in account.person.projects.all():
                if project.pid.lower() not in slurm_projects:
                    self._call([
                        "add", "user",
                        "name=%s" % username,
                        "accounts=%s" % project.pid])
        else:
            # date_deleted is not set, user should not exist
            logger.debug("account is not active")
            self._delete_account(username)

        return

    def save_account(self, account):
        """ Called when account is created/updated. """
        self._save_account(account, account.username)

    def _delete_account(self, username):
        """ Called when account is deleted. With username override. """

        # account deleted

        ds_user = self.get_user(username)
        if ds_user is not None:
            self._call(["delete", "user", "name=%s" % username])

        return

    def delete_account(self, account):
        """ Called when account is deleted. """
        self._delete_account(account.username)

    def set_account_password(self, account, raw_password):
        """ Account's password was changed. """
        pass

    def set_account_username(self, account, old_username, new_username):
        """ Account's username was changed. """
        self._delete_account(old_username)
        self._save_account(account, new_username)

    def add_account_to_project(self, account, project):
        """ Add account to project. """
        username = account.username
        projectname = project.pid
        if not self.is_user_in_project(username, projectname):
            self._call([
                "add", "user",
                "accounts=%s" % projectname,
                "name=%s" % username])

    def remove_account_from_project(self, account, project):
        """ Remove account from project. """
        username = account.username
        projectname = project.pid
        if self.is_user_in_project(username, projectname):
            self._call([
                "delete", "user",
                "name=%s" % username,
                "account=%s" % projectname])

    def account_exists(self, username):
        """ Does the account exist? """
        ds_user = self.get_user(username)
        return ds_user is not None

    def get_account_details(self, account):
        """ Get the account details """
        result = self.get_user(account.username)
        if result is None:
            result = {}
        return result

    def save_group(self, group):
        """ Group was saved. """
        pass

    def delete_group(self, group):
        """ Group was deleted. """
        pass

    def set_group_name(self, group, old_name, new_name):
        """ Group was renamed. """
        pass

    def save_project(self, project):
        """ Called when project is saved/updated. """
        pid = project.pid

        # project created
        # project updated

        if project.is_active:
            # project is not deleted
            logger.debug("project is active")
            ds_project = self.get_project(pid)
            if ds_project is None:
                self._call(["add", "account", "name=%s" % pid] + self._add_account_extra)

            # update project meta information
            name = self._truncate(project.name, 40)
            self._call([
                "modify", "account",
                "set", "Description=%s" % self._filter_string(name),
                "where", "name=%s" % pid])
            self._call([
                "modify", "account",
                "set", "Organization=%s"
                % self._filter_string(project.institute.name),
                "where", "name=%s" % pid])
        else:
            # project is deleted
            logger.debug("project is not active")
            ds_project = self.get_project(pid)
            if ds_project is not None:
                self._call(["delete", "account", "name=%s" % pid])

        return

    def delete_project(self, project):
        """ Called when project is deleted. """
        pid = project.pid

        # project deleted

        ds_project = self.get_project(pid)
        if ds_project is not None:
            self._call(["delete", "account", "name=%s" % pid])

        return

    def get_project_details(self, project):
        """ Get the project details. """
        result = self.get_project(project.pid)
        if result is None:
            result = {}
        return result

    def set_project_pid(self, project, old_pid, new_pid):
        """ Project's pid was changed. """
        # FIXME
        return


trace.attach(trace.trace(logger), SlurmDataStore)



# Copyright 2010-2017, The University of Melbourne
# Copyright 2010-2017, Brian May
#
# This file is part of Karaage.
#
# Karaage is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Karaage is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Karaage  If not, see <http://www.gnu.org/licenses/>.

import unittest

import mock
import pytest
from django.core import exceptions as django_exceptions

from karaage.projects.models import Project
from karaage.tests.fixtures import (
    GroupFactory,
    InstituteFactory,
    ProjectFactory,
    simple_account,
)
from karaage.tests.unit import UnitTestCase


@pytest.mark.django_db
class ProjectTestCase(UnitTestCase):

    def test_minimum_create(self):
        institute = InstituteFactory()
        project = Project.objects.create(
            pid='test',
            name='Test',
            institute=institute)

        project.full_clean()
        self.assertEqual(project.name, 'Test')
        self.assertEqual(project.pid, 'test')
        self.assertEqual(project.institute, institute)
        self.assertEqual(project.group.name, 'test')
        self.assertFalse(project.is_approved)
        self.assertEqual(project.leaders.count(), 0)
        self.assertTrue(project.description is None)
        self.assertTrue(project.deleted_by is None)
        self.assertTrue(project.date_deleted is None)
        self.assertTrue(project.approved_by is None)
        self.assertTrue(project.date_approved is None)
        self.assertTrue(project.last_usage is None)
        self.assertTrue(project.additional_req is None)

    @unittest.skip("broken with mysql/postgresql")
    def test_pid(self):
        assert_raises = self.assertRaises(django_exceptions.ValidationError)

        # Max length
        person = ProjectFactory(pid="a" * 255)
        person.full_clean()

        # Name is too long
        person = ProjectFactory(pid="a" * 256)
        with assert_raises:
            person.full_clean()

    def test_change_group(self):
        """Check that when changing an projects group, old accounts are
        removed from the project and new ones are added.

        """
        account1 = simple_account()
        group1 = GroupFactory()
        group1.add_person(account1.person)
        institute = InstituteFactory()

        # Test during initial creation of the project
        self.resetDatastore()
        project = Project.objects.create(group=group1, institute=institute)
        self.assertEqual(
            self.datastore.method_calls,
            [mock.call.save_project(project),
             mock.call.add_account_to_project(account1, project)])

        # Test changing an existing projects group
        account2 = simple_account()
        self.resetDatastore()
        group2 = GroupFactory()
        group2.add_person(account2.person)
        project.group = group2
        project.save()
        self.assertEqual(
            self.datastore.method_calls,
            [mock.call.save_group(group2),
             mock.call.add_account_to_group(account2, group2),
             mock.call.save_project(project),
             # old accounts are removed
             mock.call.remove_account_from_project(account1, project),
             # new accounts are added
             mock.call.add_account_to_project(account2, project)])

        # Test deleting project
        self.resetDatastore()
        project.delete()
        self.assertEqual(
            self.datastore.method_calls,
            [mock.call.remove_account_from_project(account2, project),
             mock.call.delete_project(project)])



# Copyright 2010-2017, The University of Melbourne
# Copyright 2010-2017, Brian May
#
# This file is part of Karaage.
#
# Karaage is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Karaage is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Karaage  If not, see <http://www.gnu.org/licenses/>.

import six
from django import template
from django.template import Library


register = Library()


@register.inclusion_tag('karaage/common/inlineformfield.html')
def inlineformfield(field1, field2, field3=False):
    return locals()


@register.inclusion_tag('karaage/common/form_as_div.html')
def form_as_div(form):
    return {'form': form, }


@register.tag
def formfield(parser, token):
    try:
        tag_name, field = token.split_contents()
    except ValueError:
        raise template.TemplateSyntaxError(
            "%r tag requires exactly one argument" % token.contents.split()[0])
    return FormFieldNode(field)


class FormFieldNode(template.Node):

    def __init__(self, field):
        self.field = template.Variable(field)

    def get_template(self, class_name):
        try:
            template_name = 'formfield/%s.html' % class_name
            return template.loader.get_template(template_name)
        except template.TemplateDoesNotExist:
            return template.loader.get_template('formfield/default.html')

    def render(self, context):
        try:
            field = self.field.resolve(context)
        except template.VariableDoesNotExist:
            return ''

        label_class_names = []
        if field.field.required:
            label_class_names.append('required')

        widget_class_name = field.field.widget.__class__.__name__.lower()
        if widget_class_name == 'checkboxinput':
            label_class_names.append('vCheckboxLabel')

        class_str = label_class_names and \
            six.u(' ').join(label_class_names) or six.u('')

        context.push()
        context.push()
        context['class'] = class_str
        context['formfield'] = field
        output = self.get_template(
            field.field.widget.__class__.__name__.lower()).render(
                context.flatten())
        context.pop()
        context.pop()
        return output



# -*- coding: utf-8 -*-
# Generated by Django 1.11.10 on 2018-03-01 17:21
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('kgapplications', '0003_remove_projectapplication_machine_categories'),
    ]

    operations = [
        migrations.AlterField(
            model_name='applicant',
            name='institute',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to='karaage.Institute'),
        ),
    ]



# Copyright 2010-2017, The University of Melbourne
# Copyright 2010-2017, Brian May
#
# This file is part of Karaage.
#
# Karaage is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Karaage is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Karaage  If not, see <http://www.gnu.org/licenses/>.

""" MOAB Account Manager datastore. """

import csv
import logging
import os.path
import subprocess
import sys

import karaage.common.trace as trace
from karaage.datastores import base


logger = logging.getLogger(__name__)

if sys.version_info < (3, 0):
    # Python2: csv module does not support unicode, we must use byte strings.

    def _input_csv(csv_data):
        for line in csv_data:
            assert isinstance(line, bytes)
            yield line

    def _output_csv(csv_line):
        for i, column in enumerate(csv_line):
            csv_line[i] = column.decode("ascii", errors='ignore')
            assert isinstance(csv_line[i], unicode)  # NOQA

else:
    # Python3: csv module does support unicode, we must use strings everywhere,
    # not byte strings

    def _input_csv(unicode_csv_data):
        for line in unicode_csv_data:
            assert isinstance(line, bytes)
            line = line.decode("ascii", errors='ignore')
            assert isinstance(line, str)
            yield line

    def _output_csv(csv_line):
        for column in csv_line:
            assert isinstance(column, str)


class MamDataStoreBase(base.DataStore):
    """ MAM datastore. """
    version = None

    def __init__(self, config):
        super(MamDataStoreBase, self).__init__(config)
        self._prefix = config.get('PREFIX', [])
        self._path = config.get('PATH', "/usr/local/mam/bin")
        self._null_project = config.get('NULL_PROJECT', "default")

    @staticmethod
    def _filter_string(value):
        """ Filter the string so MAM doesn't have heart failure."""
        if value is None:
            value = ""

        # replace whitespace with space
        value = value.replace("\n", " ")
        value = value.replace("\t", " ")

        # CSV seperator
        value = value.replace("|", " ")

        # remove leading/trailing whitespace
        value = value.strip()

        # hack because MAM doesn't quote sql correctly
        value = value.replace("\\", "")

        # Used for stripping non-ascii characters
        value = ''.join(c for c in value if 31 < ord(c) < 127)

        return value

    @staticmethod
    def _truncate(value, arg):
        """
        Truncates a string after a given number of chars
        Argument: Number of chars to _truncate after
        """
        length = int(arg)
        if value is None:
            value = ""
        if len(value) > length:
            return value[:length] + "..."
        else:
            return value

    def _get_command(self, command):
        path = None
        split_path = self._path.split(":")
        for d in split_path:
            tmp_path = os.path.join(d, command[0])

            if len(self._prefix) == 0:
                # short cut process if no prefix
                if os.path.isfile(tmp_path) and os.access(tmp_path, os.X_OK):
                    path = tmp_path
                    break
            else:
                # don't short cut process
                cmd = []
                cmd.extend(self._prefix)
                cmd.extend(["test", "-x", tmp_path])
                result = subprocess.call(cmd)
                if result == 0:
                    path = tmp_path
                    break
        if path is None:
            raise RuntimeError(
                "Cannot find %s in %s" % (command[0], split_path))

        cmd = []
        cmd.extend(self._prefix)
        cmd.append(path)
        cmd.extend(command[1:])
        logger.debug("Cmd %s" % command)
        return cmd

    def _call(self, command, ignore_errors=None):
        """ Call remote command with logging. """
        if ignore_errors is None:
            ignore_errors = []

        command = self._get_command(command)
        logger.debug("Cmd %s" % command)
        null = open('/dev/null', 'w')
        retcode = subprocess.call(command, stdout=null, stderr=null)
        null.close()

        if retcode in ignore_errors:
            logger.debug(
                "<-- Cmd %s returned %d (ignored)" % (command, retcode))
            return

        if retcode:
            logger.error(
                "<-- Cmd %s returned: %d (error)" % (command, retcode))
            raise subprocess.CalledProcessError(retcode, command)

        logger.debug("<-- Returned %d (good)" % retcode)
        return

    def _read_output(self, command):
        """ Read CSV delimited input from MAM. """
        command = self._get_command(command)
        null = open('/dev/null', 'w')
        process = subprocess.Popen(
            command, stdout=subprocess.PIPE, stderr=null)
        null.close()

        results = []
        reader = csv.reader(_input_csv(process.stdout), delimiter="|")

        try:
            headers = next(reader)
            logger.debug("<-- headers %s" % headers)
        except StopIteration:
            logger.debug("Cmd %s headers not found" % command)
            headers = []

        for row in reader:
            _output_csv(row)

            logger.debug("<-- row %s" % row)
            this_row = {}

            i = 0
            for i in range(0, len(headers)):
                key = headers[i]
                value = row[i]
                this_row[key] = value

            results.append(this_row)

        process.stdout.close()
        retcode = process.wait()
        if retcode != 0:
            logger.error("<-- Cmd %s returned %d (error)" % (command, retcode))
            raise subprocess.CalledProcessError(retcode, command)

        if len(headers) == 0:
            logger.debug("Cmd %s didn't return any headers." % command)

        logger.debug("<-- Returned: %d (good)" % retcode)
        return results

    def check_version(self):
        assert self.version is not None

        cmd = ["goldsh", "System", "Query", "--raw"]
        results = self._read_output(cmd)

        for row in results:
            if row['Name'] != "Moab Accounting Manager":
                continue
            if row['Version'] == self.version:
                logger.info("Found MAM version %s" % self.version)
                return True
            else:
                return False

    def get_user(self, username):
        """ Get the user details from MAM. """
        cmd = ["glsuser", "-u", username, "--raw"]
        results = self._read_output(cmd)

        if len(results) == 0:
            return None
        elif len(results) > 1:
            logger.error(
                "Command returned multiple results for '%s'." % username)
            raise RuntimeError(
                "Command returned multiple results for '%s'." % username)

        the_result = results[0]
        the_name = the_result["Name"]
        if username.lower() != the_name.lower():
            logger.error(
                "We expected username '%s' but got username '%s'."
                % (username, the_name))
            raise RuntimeError(
                "We expected username '%s' but got username '%s'."
                % (username, the_name))

        return the_result

    def get_user_balance(self, username):
        """ Get the user balance details from MAM. """
        cmd = ["gbalance", "-u", username, "--raw"]
        results = self._read_output(cmd)

        if len(results) == 0:
            return None

        return results

    def _get_project_cmd(self, projectname):
        raise NotImplementedError()

    def get_project(self, projectname):
        """ Get the project details from MAM. """
        cmd = self._get_project_cmd(projectname)
        results = self._read_output(cmd)

        if len(results) == 0:
            return None
        elif len(results) > 1:
            logger.error(
                "Command returned multiple results for '%s'." % projectname)
            raise RuntimeError(
                "Command returned multiple results for '%s'." % projectname)

        the_result = results[0]
        the_project = the_result["Name"]
        if projectname.lower() != the_project.lower():
            logger.error(
                "We expected projectname '%s' "
                "but got projectname '%s'." % (projectname, the_project))
            raise RuntimeError(
                "We expected projectname '%s' "
                "but got projectname '%s'." % (projectname, the_project))

        return the_result

    def get_users_in_project(self, projectname):
        """ Get list of users in project from MAM. """
        ds_project = self.get_project(projectname)
        if ds_project is None:
            logger.error(
                "Project '%s' does not exist in MAM" % projectname)
            raise RuntimeError(
                "Project '%s' does not exist in MAM" % projectname)

        user_list = []
        if ds_project["Users"] != "":
            user_list = ds_project["Users"].lower().split(",")
        return user_list

    def get_projects_in_user(self, username):
        """ Get list of projects in user from MAM. """
        ds_balance = self.get_user_balance(username)
        if ds_balance is None:
            return []

        project_list = []
        for bal in ds_balance:
            project_list.append(bal["Name"])
        return project_list

    def _create_user(self, username, default_project_name):
        raise NotImplementedError()

    def _set_user_default_project(self, username, default_project_name):
        raise NotImplementedError()

    def _save_account(self, account, username):
        """ Called when account is created/updated. With username override. """

        # retrieve default project, or use null project if none
        default_project_name = self._null_project
        if account.default_project is not None:
            default_project_name = account.default_project.pid

        # account created
        # account updated

        ds_user = self.get_user(username)
        if account.date_deleted is None:
            # date_deleted is not set, user should exist
            logger.debug("account is active")

            if ds_user is None:
                # create user if doesn't exist
                self._create_user(username, default_project_name)
            else:
                # or just set default project
                self._set_user_default_project(username, default_project_name)

            # update user meta information
            self._call([
                "gchuser",
                "-n", self._filter_string(account.person.get_full_name()),
                "-u", username])
            self._call([
                "gchuser",
                "-E", self._filter_string(account.person.email),
                "-u", username])

            # add rest of projects user belongs to
            mam_projects = self.get_projects_in_user(username)
            for project in account.person.projects.all():
                if project.pid not in mam_projects:
                    self.add_account_to_project(account, project)
        else:
            # date_deleted is not set, user should not exist
            logger.debug("account is not active")
            if ds_user is not None:
                # delete MAM user if account marked as deleted
                self._call(["grmuser", "-u", username], ignore_errors=[8])

        return

    def save_account(self, account):
        """ Called when account is created/updated. """
        self._save_account(account, account.username)

    def _delete_account(self, account, username):
        """ Called when account is deleted. With username override. """

        # account deleted

        ds_user = self.get_user(username)
        if ds_user is not None:
            self._call(["grmuser", "-u", username], ignore_errors=[8])

        return

    def delete_account(self, account):
        """ Called when account is deleted. """
        self._delete_account(account, account.username)

    def set_account_password(self, account, raw_password):
        """ Account's password was changed. """
        pass

    def set_account_username(self, account, old_username, new_username):
        """ Account's username was changed. """
        self._delete_account(account, old_username)
        self._save_account(account, new_username)

    def add_account_to_project(self, account, project):
        """ Add account to project. """
        raise NotImplementedError()

    def remove_account_from_project(self, account, project):
        """ Remove account from project. """
        raise NotImplementedError()

    def account_exists(self, username):
        """ Does the account exist? """
        ds_user = self.get_user(username)
        return ds_user is not None

    def get_account_details(self, account):
        """ Get the account details """
        result = self.get_user(account.username)
        if result is None:
            result = {}
        return result

    def save_group(self, group):
        """ Group was saved. """
        pass

    def delete_group(self, group):
        """ Group was deleted. """
        pass

    def set_group_name(self, group, old_name, new_name):
        """ Group was renamed. """
        pass

    def _create_project(self, pid):
        raise NotImplementedError()

    def _set_project(self, pid, description, institute):
        raise NotImplementedError()

    def _delete_project(self, pid):
        raise NotImplementedError()

    def save_project(self, project):
        """ Called when project is saved/updated. """
        pid = project.pid

        # project created
        # project updated

        if project.is_active:
            # project is not deleted
            logger.debug("project is active")
            ds_project = self.get_project(pid)
            if ds_project is None:
                self._create_project(pid)

            # update project meta information
            name = self._truncate(project.name, 40)
            self._set_project(pid, name, project.institute)
        else:
            # project is deleted
            logger.debug("project is not active")
            ds_project = self.get_project(pid)
            if ds_project is not None:
                self._delete_project(pid)

        return

    def delete_project(self, project):
        """ Called when project is deleted. """
        pid = project.pid

        # project deleted

        ds_project = self.get_project(pid)
        if ds_project is not None:
            self._delete_project(pid)

        return

    def get_project_details(self, project):
        """ Get the project details. """
        result = self.get_project(project.pid)
        if result is None:
            result = {}
        return result

    def set_project_pid(self, project, old_pid, new_pid):
        """ Project's pid was changed. """
        # FIXME
        return

    def save_institute(self, institute):
        """ Called when institute is created/updated. """
        name = institute.name
        logger.debug("save_institute '%s'" % name)

        # institute created
        # institute updated

        if institute.is_active:
            # date_deleted is not set, user should exist
            logger.debug("institute is active")

            self._call(
                ["goldsh", "Organization", "Create", "Name=%s" % name],
                ignore_errors=[185])
        else:
            # date_deleted is not set, user should not exist
            logger.debug("institute is not active")
            # delete MAM organisation if institute marked as deleted
            self._call(["goldsh", "Organization", "Delete", "Name==%s" % name])

        logger.debug("returning")
        return

    def delete_institute(self, institute):
        """ Called when institute is deleted. """
        name = institute.name
        logger.debug("institute_deleted '%s'" % name)

        # institute deleted
        self._call(["goldsh", "Organization", "Delete", "Name==%s" % name])

        logger.debug("returning")
        return


class MamDataStore71(MamDataStoreBase):
    version = "7.1"

    def _get_project_cmd(self, projectname):
        cmd = ["glsproject", "-p", projectname, "--raw"]
        return cmd

    def _create_user(self, username, default_project_name):
        self._call([
            "gmkuser", "-A",
            "-p", default_project_name,
            "-u", username])

    def _set_user_default_project(self, username, default_project_name):
        self._call([
            "gchuser",
            "-p", default_project_name,
            "-u", username])

    def add_account_to_project(self, account, project):
        """ Add account to project. """
        username = account.username
        projectname = project.pid
        self._call([
            "gchproject",
            "--add-user", username,
            "-p", projectname],
            ignore_errors=[74])

    def remove_account_from_project(self, account, project):
        """ Remove account from project. """
        username = account.username
        projectname = project.pid
        self._call([
            "gchproject",
            "--del-users", username,
            "-p", projectname])

    def _create_project(self, pid):
        self._call(["gmkproject", "-p", pid, "-u", "MEMBERS"])

    def _set_project(self, pid, description, institute):
        self._call([
            "gchproject",
            "-d", self._filter_string(description),
            "-p", pid])
        self._call([
            "gchproject",
            "-X", "Organization=%s"
            % self._filter_string(institute.name),
            "-p", pid])

    def _delete_project(self, pid):
        self._call(["grmproject", "-p", pid])


class MamDataStore72(MamDataStoreBase):
    version = "7.2"

    def _get_project_cmd(self, projectname):
        cmd = ["glsaccount", "-a", projectname, "--raw"]
        return cmd

    def _create_user(self, username, default_project_name):
        self._call([
            "gmkuser", "-A",
            "-a", default_project_name,
            "-u", username])

    def _set_user_default_project(self, username, default_project_name):
        self._call([
            "gchuser",
            "-a", default_project_name,
            "-u", username])

    def add_account_to_project(self, account, project):
        """ Add account to project. """
        username = account.username
        projectname = project.pid
        self._call([
            "gchaccount",
            "--add-user", username,
            "-a", projectname],
            ignore_errors=[74])

    def remove_account_from_project(self, account, project):
        """ Remove account from project. """
        username = account.username
        projectname = project.pid
        self._call([
            "gchaccount",
            "--del-users", username,
            "-a", projectname])

    def _create_project(self, pid):
        self._call(["gmkaccount", "-a", pid, "-u", "MEMBERS"])

    def _set_project(self, pid, description, institute):
        self._call([
            "gchaccount",
            "-d", self._filter_string(description),
            "-a", pid])
        self._call([
            "gchaccount",
            "-X", "Organization=%s"
            % self._filter_string(institute.name),
            "-a", pid])

    def _delete_project(self, pid):
        self._call(["grmaccount", "-a", pid])


def MamDataStore(config):
    ds = MamDataStore71(config)
    if ds.check_version():
        return ds
    ds = MamDataStore72(config)
    if ds.check_version():
        return ds
    raise RuntimeError("We do not support this version of MAM")


trace.attach(trace.trace(logger), MamDataStoreBase)
trace.attach(trace.trace(logger), MamDataStore71)
trace.attach(trace.trace(logger), MamDataStore72)
